doit
CoreTestCase subclass: 'CHeaderTestCase'
  instVarNames: #()
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: UserGlobals
  options: #()

%

! Remove existing behavior from CHeaderTestCase
doit
CHeaderTestCase removeAllMethods.
CHeaderTestCase class removeAllMethods.
%
! ------------------- Class methods for CHeaderTestCase
category: 'other'
set compile_env: 0
classmethod: CHeaderTestCase
wantsToBeRunByItself

	^(System gemVersionReport at: 'gsBuildArchitecture') ~= 'x86-64 (Linux)'.
%
! ------------------- Instance methods for CHeaderTestCase
category: 'utilities'
set compile_env: 0
method: CHeaderTestCase
assertType: aSymbol fromString: aString

	| header types declaration |
	self
		assert: (header := self string: aString) notNil;
		assert: (declaration := (types := header types) at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		assert: declaration storage equal: #'typedef';
		assert: declaration type equal: aSymbol;
		yourself.
%
category: 'utilities'
set compile_env: 0
method: CHeaderTestCase
removeServerFile: aString

	(GsFile existsOnServer: aString) ifTrue: [
		(GsFile removeServerFile: aString) isNil ifTrue: [
			self error: GsFile serverErrorString.
		].
	].
%
category: 'utilities'
set compile_env: 0
method: CHeaderTestCase
sample
"
UserGlobals at: #'James' put: (CPreprocessor parseFileAt: '$GEMSTONE/include/gci.hf').
James tokens size.
CHeader tokens: James tokens.
"
| stream start list |
stream := WriteStream on: String new.
start := 12550.
list := (UserGlobals at: #'James') tokens copyFrom: start to: start + 100.
list do: [:each | 
	stream nextPutAll: each value asString.
	each = (#'punctuator' -> ';') 
		ifTrue: [stream cr]
		ifFalse: [stream space].
].
^stream contents.
%
category: 'utilities'
set compile_env: 0
method: CHeaderTestCase
string: string

	^CHeader string: string.
%
category: 'other'
set compile_env: 0
method: CHeaderTestCase
testAAA

	| string header "declaration parameters" |
	string := '
'.
	self
		assert: (header := self string: string) notNil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testAfterEmpty

	| string header storage data |
	string := '
#define foo
foo
a;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (storage := header storage) size equal: 1;
		assert: (data := storage at: 'a') notNil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testBitFields

	| string header declaration struct fields |
string := '
struct { 
	unsigned int a : 7 ;
	unsigned int b : 1 ;
	unsigned int c : 8 ;
	unsigned int : 16 ;
} foo ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'foo') notNil;
		assert: declaration baseByteSize equal: 4;
		assert: (struct := declaration type) class equal: CDeclaration;
		assert: (fields := struct fields) size equal: 4;
		assert: (declaration := fields at: 1) notNil;
		assert: declaration name equal: 'a';
		assert: declaration bitCount equal: 7;
		assert: (declaration := fields at: 4) notNil;
		assert: declaration name equal: nil;
		assert: declaration bitCount equal: 16;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testCalculatedSize

	| string header declaration fields |
string := '
typedef struct { 
	unsigned long int __val [ ( 1024 / ( 8 * sizeof ( unsigned long int ) ) ) ] ;
} __sigset_t ;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: '__sigset_t') notNil;
		assert: declaration storage equal: #'typedef';
		assert: (declaration := declaration type) class equal: CDeclaration;
		assert: declaration type equal: #'struct';
		assert: (fields := declaration fields) size equal: 1;
		assert: (declaration := fields first) notNil;
		assert: declaration name equal: '__val';
		assert: declaration type equal: #'uint64';
		assert: declaration count equal: #(16);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testCastInExpression

	| string header declaration fields |
string := '
typedef long int __fd_mask ;
typedef struct { 
	__fd_mask __fds_bits [ 1024 / ( 8 * ( int ) sizeof ( __fd_mask ) ) ] ;
} fd_set ;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: '__fd_mask') notNil;
		assert: declaration storage equal: #'typedef';
		assert: declaration type equal: #'int64';

		assert: (declaration := header types at: 'fd_set') notNil;
		assert: declaration storage equal: #'typedef';
		assert: (declaration := declaration type) class equal: CDeclaration;
		assert: declaration type equal: #'struct';
		assert: (fields := declaration fields) size equal: 1;
		assert: (declaration := fields first) notNil;
		assert: declaration name equal: '__fds_bits';
		assert: declaration type equal: #'int64';
		assert: declaration count equal: #(16);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testCharArray

	| string header function parameter |
	string := '
int foo(const char bar[]);
'.
	self
		assert: (header := self string: string) notNil;
		assert: (function := header functions at: 'foo') notNil;
		assert: (parameter := function parameters at: 1) notNil;
		assert: parameter argumentType equal: #'const char*';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testClassReferenceInClass

	| string header declaration properties |
string := '
class foo {
	public:
		int a;
} ;
class bar {
	public:
		foo myFoo;
} ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: (properties := declaration properties) size equal: 1;
		assert: (declaration := header types at: 'bar') notNil;
		assert: (properties := declaration properties) size equal: 1;
		assert: (declaration := properties at: 1) notNil;
		assert: declaration type equal: #'class';
		assert: declaration name equal: 'myFoo';
		assert: (properties := declaration properties) size equal: 1;
		assert: (declaration := properties at: 1) notNil;
		assert: declaration type equal: #'int32';
		assert: declaration name equal: 'a';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testCommentMacroInDeclaration

	| string header declaration |
string := '
#define __wur  /* ignore */
extern int foo ( int * x , int * y ) __attribute__ ( ( __nothrow__ ) ) __attribute__ ( ( __nonnull__ ( 2 ) ) ) __wur ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testConstArgumentType

	| string header function parameters parameter |
	string := 'extern int foo ( __const char *foo, const char* bar ) ;'.
	self
		assert: (header := self string: string) notNil;
		assert: (function := header functions at: 'foo') isFunction;
		assert: (parameters := function parameters) size equal: 2;
		assert: (parameter := parameters at: 1) name equal: 'foo';
		assert: parameter isConstant;
		assert: (parameter := parameters at: 2) name equal: 'bar';
		assert: parameter isConstant;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testConstructorFunctionInClass

	| string header declaration |
string := '
class foo {
	public:
		foo (size_t bar ) ;
}
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration properties equal: #();
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testConstructorFunctionWithDefaults

	| string header declaration |
string := '
#define NULL ((void *) 0)
class foo {
	public:
		foo ( int x = 1, const char * y = NULL ) ;
}
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration properties equal: #();
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testDestructorFunction

	| string header declaration |
string := '
class foo {
	public:
		~ foo ( ) ;
}
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration properties equal: #();
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testEmpty_struct
"
"
	| string header types structs type fields storage data |
	string := '
struct _IO_FILE; 
typedef struct _IO_FILE FILE;
extern struct _IO_FILE *stdin;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (structs := header structs) size equal: 1;
		assert: (fields := header fieldsForStruct: '_IO_FILE') isNil;
		assert: (types := header types) size equal: 1;
		assert: (type := types at: 'FILE') notNil;
		assert: type name equal: 'FILE';
		assert: type storage equal: #'typedef';
		assert: (type := type type) class equal: CDeclaration;
		assert: type type equal: #'struct';
		assert: type fields equal: fields;

		assert: (storage := header storage) notNil;
		assert: (storage includesKey: 'stdin');
		assert: (data := storage at: 'stdin') notNil;
		assert: data name equal: 'stdin';
		assert: data storage equal: #'extern';
		assert: data type identical: type;
		assert: data pointer equal: 1;
		assert: data fields equal: fields;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testEnumAsArraySize

	| string header declaration |
	string := '
enum { FOO =  10 };
int foo[FOO];
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'foo') notNil;
		assert: declaration count equal: #(10);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testExpressionPrecedence

	| string header |
string := '
char a [ 1 + 2 + 3 ];
char b [ 1 * 2 + 3 ];
char c [ 1 + 2 * 3 ];
char d [ 5 * 4 - 3 * 2 ];
char e [ 5 * 4 + 3 * 2 + 1 ];
char f [ 20 - 3 * 2 + 1 ];
'.
	self
		assert: (header := self string: string) notNil;
		assert: (header storage at: 'a') count equal: #(6);
		assert: (header storage at: 'b') count equal: #(5);
		assert: (header storage at: 'c') count equal: #(7);
		assert: (header storage at: 'd') count equal: #(14);
		assert: (header storage at: 'e') count equal: #(27);
		assert: (header storage at: 'f') count equal: #(15);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testFunction

	| string header function parameters parameter |
	string := 'extern void foo ( foo );'.
	self
		assert: (header := self string: string) notNil;
		assert: header types isEmpty;
		assert: header structs isEmpty;
		assert: header storage isEmpty;
		assert: (function := header functions at: 'foo') isFunction;
		assert: function name equal: 'foo';
		assert: function storage equal: #'extern';
		assert: function type equal: #'void';
		assert: (parameters := function parameters) size equal: 1;
		assert: (parameter := parameters first) notNil;
		assert: parameter storage equal: #'auto';
		assert: parameter type equal: #'int32';
		assert: parameter name equal: 'foo';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testFunctionNoArguments

	| string header function parameters |
	string := 'extern void foo ( void );'.
	self
		assert: (header := self string: string) notNil;
		assert: header types isEmpty;
		assert: header structs isEmpty;
		assert: header storage isEmpty;
		assert: (function := header functions at: 'foo') isFunction;
		assert: function name equal: 'foo';
		assert: function storage equal: #'extern';
		assert: function type equal: #'void';
		assert: (parameters := function parameters) size equal: 0;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testFunctionReturningFunction

	| string header declaration parameters |
	string := '
//	define a function, zork, that takes a long and returns a function that takes a short and returns an int
extern int (* zork( long foo ) )( short bar );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'zork') name equal: 'zork';
		assert: (parameters := declaration parameters) size equal: 1;
		assert: parameters first type equal: #int64;
		assert: (declaration := declaration type) name equal: nil;
		assert: (parameters := declaration parameters) size equal: 1;
		assert: parameters first type equal: #int16;
		assert: declaration type equal: #int32;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testFunctionSource

	| string1 string2 string header functions function |
	string1 := '
# define __MALLOC_P(args)	args __THROW
# define __attribute_malloc__ __attribute__ ((__malloc__))
# define __wur /* Ignore */'.
	string2 := '
/* Allocate SIZE bytes of memory.  */
extern void *malloc __MALLOC_P ((size_t __size)) __attribute_malloc__ __wur;'.
	string := string1 , string2.
	self
		assert: (header := self string: string) notNil;
		assert: (functions := header functions) size equal: 1;
		assert: (function := functions at: 'malloc') notNil;
		assert: function source equal: string2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testFunctionTwoArguments

	| string header function parameters parameter |
	string := 'extern void foo ( int a, int b );'.
	self
		assert: (header := self string: string) notNil;
		assert: header types isEmpty;
		assert: header structs isEmpty;
		assert: header storage isEmpty;
		assert: (function := header functions at: 'foo') isFunction;
		assert: function name equal: 'foo';
		assert: function storage equal: #'extern';
		assert: function type equal: #'void';
		assert: (parameters := function parameters) size equal: 2;
		assert: (parameter := parameters at: 1) name equal: 'a';
		assert: (parameter := parameters at: 2) name equal: 'b';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testFunctionTypeAsReturnAndArgument

	| string header foo bar |
	string := '
typedef void foo(void);
typedef foo * bar(foo in, foo *out);
'.
	self
		assert: (header := self string: string) notNil;
		assert: (foo := header types at: 'foo') notNil;
		assert: foo isFunction;
		assert: foo type equal: #'void';
		assert: (bar := header types at: 'bar') notNil;
		assert: bar isFunction;
		assert: bar type equal: #'function';
		assert: bar pointer equal: 1;
		assert: bar parameters size equal: 2;
		assert: (foo := bar parameters at: 1) notNil;
		assert: foo type equal: #'function';
		assert: foo pointer equal: 0;
		assert: (foo := bar parameters at: 2) notNil;
		assert: foo type equal: #'function';
		assert: foo pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testFunctionWithFunctionPointerArgument

	| string header declaration parameters |
string := '
extern int foo ( int ( * bar ) ( unsigned int ) ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		assert: declaration isFunction;
		deny: declaration isStorage;
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters at: 1) notNil;
		assert: declaration isFunction;
		assert: declaration isStorage;
		assert: declaration name equal: 'bar';
		assert: declaration type equal: #'int32';
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters at: 1) notNil;
		assert: declaration name equal: nil;
		assert: declaration type equal: #'uint32';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testMultipleAttributes

	| string header declaration parameters |
string := '
extern void foo ( int * bar ) __attribute__ ( ( __nothrow__ ) ) __attribute__ ( ( __nonnull__ ( 1 ) ) ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testMultipleTypedefsOnOneLine

	| string header declaration |
	string := '
typedef int *foo, *bar;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration type equal: #int32;
		assert: declaration pointer equal: 1;
		assert: (declaration := header types at: 'bar') notNil;
		assert: declaration type equal: #int32;
		assert: declaration pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testMultipleVariablesPerLine

	| string header foo bar |
	string := '
 int **foo, *bar;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (foo := header storage at: 'foo') name equal: 'foo';
		assert: foo pointer equal: 2;
		assert: (bar := header storage at: 'bar') name equal: 'bar';
		assert: bar pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testMultipleVariablesPerLineInStruct

	| string header typedef struct fields foo bar |
	string := '
typedef struct {
	int foo, *bar;
} Visual'.
	self
		assert: (header := self string: string) notNil;
		assert: (typedef := header types at: 'Visual') notNil;
		assert: (struct := typedef type) notNil;
		assert: (fields := struct fields) size equal: 2;
		assert: (foo := fields at: 1) name equal: 'foo';
		assert: (bar := fields at: 2) name equal: 'bar';
		assert: bar pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testNegative_enum

	| string header  |
string := '
enum __rusage_who { RUSAGE_SELF = 0 , RUSAGE_CHILDREN = - 1 } ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (header enums at: 'RUSAGE_SELF') = 0;
		assert: (header enums at: 'RUSAGE_CHILDREN') = -1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testParenthesisExpression

	| string header declarations declaration |
string := '
#if ((!defined __STRICT_ANSI__ || (_XOPEN_SOURCE - 0) >= 500) && \
     !defined _POSIX_SOURCE && !defined _POSIX_C_SOURCE)
aaa
#else
bbb
#endif
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declarations := header declarations) size equal: 1;
		assert: (declaration := declarations at: 1) notNil;
		assert: declaration name equal: 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testRecursive_struct
"
"
	| string header struct type declaration fields |
	string := '
typedef struct foo { 
	struct foo * prev ;
	struct foo * next ;
} foo_t ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (struct := header structs at: 'foo') notNil;
		assert: (fields := struct fields) size equal: 2;
		assert: (type := header types at: 'foo_t') notNil;
		assert: type type identical: struct;
		assert: type pointer equal: 0;
		assert: (declaration := fields at: 1) notNil;
		assert: declaration name equal: 'prev';
		assert: declaration type identical: struct;
		assert: declaration pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testRestrictArgumentType

	| string header function parameters parameter |
	string := 'extern int foo ( __const char * __restrict foo, const char* restrict bar ) ;'.
	self
		assert: (header := self string: string) notNil;
		assert: (function := header functions at: 'foo') isFunction;
		assert: (parameters := function parameters) size equal: 2;
		assert: (parameter := parameters at: 1) name equal: 'foo';
		assert: parameter isConstant;
		assert: (parameter := parameters at: 2) name equal: 'bar';
		assert: parameter isConstant;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testSignedWithNoSize

	| string header declaration |
	string := '
unsigned foo;
signed bar;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'foo') notNil;
		assert: declaration type equal: #uint32;
		assert: (declaration := header storage at: 'bar') notNil;
		assert: declaration type equal: #int32;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testStaticFunctionInClass

	| string header declaration properties |
string := '
class foo {
		static bar() { return sizeof(foo) }
} ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: (properties := declaration properties) size equal: 0;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testStorageOfFunctionPointer

	| string header declaration |
	string := '
extern void *(*foo) (int bar);
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'foo') notNil;
		assert: declaration pointer equal: 1;
		assert: declaration isStorage;
		assert: declaration isFunction;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testTwoFunction

	| string header |
	string := '
extern void setpwent ( int a, int b ) ;
extern void endpwent ( void ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: header functions size equal: 2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testTypeCastArraySize

	| string header declaration |
	string := '
typedef int INT;
char foo[(int) 8];
char bar[(INT) 8];
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'foo') notNil;
		assert: declaration count equal: #(8);
		assert: (declaration := header storage at: 'bar') notNil;
		assert: declaration count equal: #(8);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testUnnamedArgument

	| string header declaration parameters |
string := '
extern int foo ( int * __restrict ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters at: 1) notNil;
		assert: declaration name equal: nil;
		assert: declaration type equal: #'int32';
		assert: declaration pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testUnnamedFunctionAsArgument

	| string header declaration parameters |
	string := '
extern int (* zork( long (*)(char foo) ) )( short bar );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'zork') name equal: 'zork';
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters at: 1) isFunction;
		assert: declaration name equal: nil;
		assert: declaration type equal: #'int64';
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters at: 1) name equal: 'foo';
		assert: declaration type equal: #'uint8';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testUnnamedFunctionParameter

	| string header declaration parameters |
	string := '
extern void foo( char [32] );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') name equal: 'foo';
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters at: 1) name equal: nil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testUnnamedFunctionParameterAfterEmptyMacro

	| string header declaration parameters |
	string := '
#  define STDC
#  define FAR

#ifndef OF /* function prototypes */
#  ifdef STDC
#    define OF(args)  args
#  else
#    define OF(args)  ()
#  endif
#endif
 
typedef unsigned (*in_func) OF((void FAR *, unsigned char FAR * FAR *));
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'in_func') notNil;
		assert: declaration type equal: #'uint32';
		assert: (parameters := declaration parameters) size equal: 2;
		assert: (declaration := parameters at: 1) notNil;
		assert: declaration name equal: nil;
		assert: declaration type equal: #'void';
		assert: declaration pointer equal: 1;
		assert: (declaration := parameters at: 2) notNil;
		assert: declaration name equal: nil;
		assert: declaration type equal: #'uint8';
		assert: declaration pointer equal: 2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testVariableLengthStruct

	| string header fields declaration |
string := '
struct cmsghdr { 
	size_t cmsg_len ;
	int cmsg_level ;
	int cmsg_type ;
	__extension__ unsigned char __cmsg_data [ ] ;
} ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (fields := header fieldsForStruct: 'cmsghdr') size equal: 4;
		assert: (declaration := fields at: 1) count equal: nil;
		assert: (declaration := fields at: 4) count equal: #(0);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testVaryingArgs

	| string header declaration parameters |
string := '
extern int foo ( int bar , . . . ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		assert: declaration isVaryingArgCount;
		assert: (parameters := declaration parameters) size equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
testVoidPointerArgument

	| string header function parameters |
	string := '
void * foo ( void * __ptr ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (function := header functions at: 'foo') isFunction;
		assert: (parameters := function parameters) size equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_attribute_const

	| string header declaration parameters |
string := '
extern int foo ( int bar ) __attribute__ ( ( __const__ ) );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_attribute_malloc

	| string header declaration parameters |
string := '
extern int foo ( int bar ) __attribute__ ( ( __malloc__ ) );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_attribute_mode
"
	/usr/include/sys/types.h

	http://tigcc.ticalc.org/doc/gnuexts.html#SEC91_mode
	http://www.linuxtopia.org/online_books/programming_tool_guides/linux_using_gnu_compiler_collection/vector-extensions.html
		QI = 8-bit integer, HI = 16-bit integer, SI = 32-bit integer, DI = 64-bit integer, SF = 32-bit float, DF = 64-bit float
		byte = 8-bit integer, word = 64-bit integer, pointer = 64-bit integer
" 
	| string header types declaration |
string := '
# define __intN_t(N, MODE) \
  typedef int int##N##_t __attribute__ ((__mode__ (MODE)))
# define __u_intN_t(N, MODE) \
  typedef unsigned int u_int##N##_t __attribute__ ((__mode__ (MODE)))

# ifndef __int8_t_defined
#  define __int8_t_defined
__intN_t (8, __QI__);
__intN_t (16, __HI__);
__intN_t (32, __SI__);
__intN_t (64, __DI__);
# endif

__u_intN_t (8, __QI__);
__u_intN_t (16, __HI__);
__u_intN_t (32, __SI__);
__u_intN_t (64, __DI__);

typedef int register_t __attribute__ ((__mode__ (__word__)));
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := (types := header types) at: 'int8_t') type equal: #'int8';
		assert: (declaration := header types at: 'int16_t') type equal: #'int16';
		assert: (declaration := header types at: 'int32_t') type equal: #'int32';
		assert: (declaration := header types at: 'int64_t') type equal: #'int64';
		assert: (declaration := header types at: 'u_int8_t') type equal: #'uint8';
		assert: (declaration := header types at: 'u_int16_t') type equal: #'uint16';
		assert: (declaration := header types at: 'u_int32_t') type equal: #'uint32';
		assert: (declaration := header types at: 'u_int64_t') type equal: #'uint64';
		assert: (declaration := header types at: 'register_t') type equal: #'int64';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_attribute_sentinel

	| string header declaration parameters |
	string := '
extern void foo(void) __attribute__( ( __sentinel__( 0 ) ) );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') name equal: 'foo';
		assert: (parameters := declaration parameters) size equal: 0;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_class

	| string header declaration properties |
string := '
class foo {
	public:
		int a;
	private: 
		int b;
	public:
		enum { FOO_1 = 1 , };
		inline int x() const { return b >> 24 ; }
		static y() { return sizeof(foo) }
} ;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration source equal: string;
		assert: (properties := declaration properties) size equal: 2;
		assert: (declaration := properties at: 1) name equal: 'a';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_const
"
	We support only a subset of the 'const' possible meanings--
	basically, enought to get #'const char*' as a function argument
	type. See CCallout class>>#'library:name:result:args:'.
"
	| string header declaration |
string := '
const int a;				// data is constant (no pointer)
const int *b;				// data is constant; pointer can change
int const *c;				// data is constant; pointer can change
int * const d;				// pointer is constant; data can change
const int * const e;		// pointer is constant; data is constant

extern __const char * __const _sys_siglist [ 65 ] ;

'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'a') notNil;
		assert: declaration isConstant;
		assert: (declaration := header storage at: 'b') notNil;
		assert: declaration isConstant;
		assert: (declaration := header storage at: 'c') notNil;
		assert: declaration isConstant;
		assert: (declaration := header storage at: 'd') notNil;
		deny: declaration isConstant;
		assert: (declaration := header storage at: 'e') notNil;
		assert: declaration isConstant;
		assert: (declaration := header storage at: '_sys_siglist') notNil;
		assert: declaration isConstant;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_deprecated

	| string header declaration parameters |
string := '
extern int foo ( int * a, int * b, int * c) __attribute__ ((__deprecated__));
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 3;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_enum

	| string header  |
string := '
enum status { run, create, delete=5, suspend, killed = 9 };
'.
	self
		assert: (header := self string: string) notNil;
		assert: (header enums at: 'run') = 0;
		assert: (header enums at: 'create') = 1;
		assert: (header enums at: 'delete') = 5;
		assert: (header enums at: 'suspend') = 6;
		assert: (header enums at: 'killed') = 9;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_extension

	| string header declaration |
string := '
 __extension__ typedef int foo ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_extern_C

	| string header declaration |
string := '
int a;
extern "C" {
int b;
}
int c;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'a') notNil;
		assert: (declaration := header storage at: 'b') notNil;
		assert: (declaration := header storage at: 'c') notNil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_extern_C_function

	| string header declaration |
string := '
int a;
extern "C" {
int foo();
}
int c;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'a') notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration storage equal: #'extern';
		assert: (declaration := header storage at: 'c') notNil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_format

	| string header declaration parameters |
string := '
  extern int snprintf ( char * __restrict __s , size_t __maxlen , __const char * __restrict __format , . . . ) __attribute__ ( ( __nothrow__ ) ) __attribute__ ( ( __format__ ( __printf__ , 3 , 4 ) ) ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'snprintf') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 3;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_nonnull

	| string header declaration parameters |
string := '
extern int open ( __const char * __file , int __oflag , . . . ) __attribute__ ( ( __nonnull__ ( 1 ) ) ) ;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'open') notNil;
		assert: declaration storage equal: #'extern';
		assert: declaration type equal: #'int32';
		assert: declaration isFunction;
		assert: declaration isVaryingArgCount;
		assert: (parameters := declaration parameters) size equal: 2;
		assert: (declaration := parameters at: 2) notNil;
		assert: declaration name equal: '__oflag';
		assert: declaration type equal: #'int32';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_nonnull2

	| string header declaration parameters |
string := '
# define __nonnull(params) __attribute__ ((__nonnull__ params))
extern int foo ( int * a, int * b, int * c) __nonnull ((1, 2, 3));
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 3;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_noreturn

	| string header declaration parameters |
string := '
extern int foo ( int bar ) __attribute__ ( ( __noreturn__ ) );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_nothrow

	| string header declaration parameters |
string := '
 __extension__ extern unsigned int gnu_dev_major ( unsigned long long int __dev ) __attribute__ ( ( __nothrow__ ) ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'gnu_dev_major') notNil;
		assert: declaration storage equal: #'extern';
		assert: declaration type equal: #'uint32';
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters first) notNil;
		assert: declaration name equal: '__dev';
		assert: declaration type equal: #'uint64';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_ptrdiff_t

	| string header declaration parameters |
string := '
extern void * foo ( ptrdiff_t bar ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters at: 1) notNil;
		assert: declaration name equal: 'bar';
		assert: declaration type equal: #'int64';
		assert: declaration pointer equal: 0;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_pure

	| string header declaration parameters |
string := '
extern int foo ( int bar ) __attribute__ ( ( __pure__ ) );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_sizeof_struct

	| string header declaration |
string := '
struct foo { 
	int cmsg_level ;
	int cmsg_type ;
} ;
int bar [ sizeof( struct foo ) + sizeof( struct foo ) ] ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'bar') notNil;
		assert: declaration count equal: #(16);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_sizeof_void_pointer

	| string header declaration |
string := '
char foo [ sizeof ( void * ) ] ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'foo') notNil;
		assert: declaration count equal: #(8);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_static_inline

	| string header  |
string := '
static inline long foo ( int arg ) { return ( ( OopType ) ( ( int64 ) ( arg ) << 3 ) | 2 ) ; }
enum { GCI_MAX_SYMBOL_SIZE = 1024 } ;
'.
	self
		assert: (header := self string: string) notNil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_struct_identifier

	| string header structs fields declaration |
	string := '
struct foo { 
	int bar; // comment for bar
};'.
	self
		assert: (header := self string: string) notNil;
		assert: (header types at: 'foo' ifAbsent: [nil]) equal: nil;
		assert: (structs := header structs) notNil;
		assert: (structs isKindOf: AbstractDictionary);
		assert: (fields := header fieldsForStruct: 'foo') size equal: 1;
		assert: (declaration := fields first) notNil;
		assert: declaration name equal: 'bar';
		assert: declaration storage == #'auto';
		assert: declaration type equal: #'int32';
		assert: declaration pointer equal: 0;
		assert: declaration count equal: nil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_struct_size

	| string header declaration |
	string := '
typedef struct {
	int			a;
	long		b;
	char		c;
	char *	d;
	int			e[5 * 1][2];
} foo;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration byteSize equal: 61;
		assert: declaration source equal: string;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_transparent_union

	| string header declaration |
string := '
 typedef union { 
	long foo ;
	short bar ;
} foobar __attribute__ ( ( __transparent_union__ ) ) ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foobar') notNil;
		assert: declaration isTransparentUnion;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_char

	self
		assertType: #'uint8' 	fromString: 'typedef char foo;';
		assertType: #'uint8' 	fromString: 'typedef unsigned char foo;';
		assertType: #'int8' 		fromString: 'typedef signed char foo;';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_enum

	| string header  |
string := '
typedef enum { P_ALL , P_PID , P_PGID } idtype_t ;
'.
	self
		assert: (header := self string: string) notNil;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_float

	self
		assertType: #'float' 			fromString: 'typedef float foo;';					"32-bits"
		assertType: #'double' 			fromString: 'typedef double foo;';				"64-bits"
		assertType: #'longDouble' 	fromString: 'typedef long double foo;';		"128-bits"
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_int16

	self
		assertType: #'int16' fromString: 'typedef short foo;';
		assertType: #'int16' fromString: 'typedef short int foo;';
		assertType: #'int16' fromString: 'typedef signed short foo;';
		assertType: #'int16' fromString: 'typedef signed short int foo;';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_int32

	self
		assertType: #'int32' fromString: 'typedef int foo;';
		assertType: #'int32' fromString: 'typedef signed int foo;';

		assertType: #'int32' fromString: 'typedef wchar_t foo;';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_int64

	self
		assertType: #'int64' fromString: 'typedef long foo;';
		assertType: #'int64' fromString: 'typedef long long foo;';
		assertType: #'int64' fromString: 'typedef long int foo;';
		assertType: #'int64' fromString: 'typedef signed long foo;';
		assertType: #'int64' fromString: 'typedef signed long int foo;';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_int_star

	| string header types declaration |
	string := 'typedef int * foo;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := (types := header types) at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		assert: declaration storage equal: #'typedef';
		assert: declaration type equal: #'int32';
		assert: declaration pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_struct_int_array

	| string header types type fields |
	string := 'typedef struct { int myInt[2]; } foo;'.
	self
		assert: (header := self string: string) notNil;
		assert: header structs isEmpty;
		assert: (types := header types) size equal: 1;
		assert: (type := types at: 'foo') notNil;
		assert: type name equal: 'foo';
		assert: type storage equal: #'typedef';
		assert: (type := type type) class equal: CDeclaration;
		assert: type type equal: #'struct';
		assert: type pointer equal: 0;
		assert: (fields := type fields) size equal: 1;
		assert: (type := fields first) notNil;
		assert: type name equal: 'myInt';
		assert: type storage == #'auto';
		assert: type type equal: #'int32';
		assert: type pointer equal: 0;
		assert: type count equal: #(2);
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_typedef

	| string header types declaration |
	string := 'typedef int foo; typedef foo bar;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := (types := header types) at: 'bar') notNil;
		assert: declaration name equal: 'bar';
		assert: declaration storage equal: #'typedef';
		assert: declaration type equal: #'int32';
		assert: declaration pointer equal: 0;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_uint16

	self
		assertType: #'uint16' fromString: 'typedef unsigned short foo;';
		assertType: #'uint16' fromString: 'typedef unsigned short int foo;';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_uint32

	self
		assertType: #'uint32' fromString: 'typedef unsigned int foo;';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_uint64

	self
		assertType: #'uint64' fromString: 'typedef size_t foo;';
		assertType: #'uint64' fromString: 'typedef unsigned long foo;';
		assertType: #'uint64' fromString: 'typedef unsigned long int foo;';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_unsigned_short_star

	| string header declaration |
	string := '
typedef unsigned short * foo ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		assert: declaration storage equal: #'typedef';
		assert: declaration type equal: #'uint16';
		assert: declaration pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_void_star

	| string header declaration |
	string := 'typedef void * foo;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		assert: declaration storage equal: #'typedef';
		assert: declaration type equal: #'void';
		assert: declaration pointer equal: 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_typedef_void_star_star

	| string header declaration |
	string := 'typedef void ** foo;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		assert: declaration storage equal: #'typedef';
		assert: declaration type equal: #'void';
		assert: declaration pointer equal: 2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_union

	| string header declaration members |
string := '
typedef union { 
	char __size [ 56 ] ;
	long int __align ;
} pthread_attr_t ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header types at: 'pthread_attr_t') notNil;
		assert: (declaration := declaration type) class equal: CDeclaration;
		assert: declaration type equal: #'union';
		assert: (members := declaration members) size equal: 2;
		assert: (declaration := members at: 1) notNil;
		assert: declaration name equal: '__size';
		assert: declaration type equal: #'uint8';
		assert: declaration count equal: #(56);

		assert: (declaration := members at: 2) notNil;
		assert: declaration name equal: '__align';
		assert: declaration type equal: #'int64';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_va_list

	| string header declaration parameters |
string := '
extern int foo ( __gnuc_va_list bar ) ;'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: (parameters := declaration parameters) size equal: 1;
		assert: (declaration := parameters at: 1) notNil;
		assert: declaration name equal: 'bar';
		assert: declaration type equal: #'ptr';
		assert: declaration pointer equal: 0;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_volatile

	| string header declaration |
string := '
volatile int foo ;
typedef volatile int bar ;
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header storage at: 'foo') notNil;
		assert: declaration name equal: 'foo';
		assert: (declaration := header types at: 'bar') notNil;
		assert: declaration name equal: 'bar';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CHeaderTestCase
test_warn_unused_result

	| string header declaration parameters |
string := '
extern int foo ( int bar ) __attribute__ ( ( __warn_unused_result__ ) );
'.
	self
		assert: (header := self string: string) notNil;
		assert: (declaration := header functions at: 'foo') notNil;
		assert: declaration isFunction;
		assert: (parameters := declaration parameters) size equal: 1;
		yourself.
%
