!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!=========================================================================

doit
CoreTestCase subclass: 'CPreprocessorTestCase'
  instVarNames: #()
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: UserGlobals
  options: #()

%

! Remove existing behavior from CPreprocessorTestCase
doit
CPreprocessorTestCase removeAllMethods.
CPreprocessorTestCase class removeAllMethods.
%
! ------------------- Class methods for CPreprocessorTestCase
category: 'other'
set compile_env: 0
classmethod: CPreprocessorTestCase
bypassLogging
"
	CPreprocessorTestCase bypassLogging.
"
	System _sessionStateAt: 21 put: (WriteStream on: String new).
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorTestCase
wantsToBeRunByItself

	^(System gemVersionReport at: 'gsBuildArchitecture') ~= 'x86-64 (Linux)'.
%
! ------------------- Instance methods for CPreprocessorTestCase
category: 'utility'
set compile_env: 0
method: CPreprocessorTestCase
parseString: string

	^CPreprocessor parseString: string.
%
category: 'other'
set compile_env: 0
method: CPreprocessorTestCase
testAAA

	| string tokens |
	string := '
'.
	tokens := (self parseString: string) tokens.
	self
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testAndOperator

	| string tokens |
	string := '
#define JAMES _a1_
#if defined(JAMES) && defined(BEVERLY)
aaa
#else
bbb
#endif /* JAMES */
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testBitShift

	| string tokens x |
	string := '
#if (1 << 2) == 4 && (4 >> 1) == 2
aaa
#else
bbb
#endif'.	
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testBlockComment

	| string tokens token |
	string := 'aaa/* xxx */bbb'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 2;
		assert: (token := tokens at: 1) notNil;
		assert: token line equal: 1;
		assert: token isIdentifierToken;
		assert: token value equal: 'aaa';
		assert: token source equal: 'aaa/* xxx */';

		assert: (token := tokens at: 2) notNil;
		assert: token line equal: 1;
		assert: token isIdentifierToken;
		assert: token value equal: 'bbb';
		assert: token source equal: 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testCDefs46

	| string tokens |
	string := '
#define FOO(x) (1 < x)
#if 1
# if 1 && FOO (2)
aaa
# else
bbb
# endif
#endif
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testCharacterToken

	| string preprocessor tokens |
	string := '      *d = ''\0'';  '.
	preprocessor := self parseString: string.
	self 
		assert: (tokens := preprocessor tokens) size equal: 5;
		assert: (tokens at: 1) asAssociation equal: #'punctuator' -> '*';
		assert: (tokens at: 2) asAssociation equal: #'identifier' -> 'd';
		assert: (tokens at: 3) asAssociation equal: #'punctuator' -> '=';
		assert: (tokens at: 4) asAssociation equal: #'uint8' -> 0;
		assert: (tokens at: 5) asAssociation equal: #'punctuator' -> ';';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testCommentAfterStringLiteral

	| string tokens |
	string := 'aaa"bbb"ccc/* xxx */ddd'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 4;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'aaa';
		assert: (tokens at: 2) asAssociation equal: #'string' -> 'bbb';
		assert: (tokens at: 2) source equal: '"bbb"';
		assert: (tokens at: 3) asAssociation equal: #'identifier' -> 'ccc';
		assert: (tokens at: 4) asAssociation equal: #'identifier' -> 'ddd';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testCommentInStringLiteral

	| string tokens |
	string := 'aaa"/* xxx */"bbb'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 3;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'aaa';
		assert: (tokens at: 2) asAssociation equal: #'string' -> '/* xxx */';
		assert: (tokens at: 3) asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testConcatenation

	| preprocessor string tokens |
	string := '
#define FOO(name) foo_ ## name
FOO(bar)
'.
	preprocessor := self parseString: string.
	self
		assert: (tokens := preprocessor tokens) size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'foo_bar';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testConcatenation1

	| preprocessor string tokens |
	string := '
#define FOO(name) foo_ ## name
FOO(1)
'.
	preprocessor := self parseString: string.
	self
		assert: (tokens := preprocessor tokens) size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'foo_1';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testDecimalConstants

	| string tokens |
	string := '
0 32767 65535 2147483647 4294967295 9223372036854775807 18446744073709551615
0l 0L 1UL 1LU 2U 2u
'.
	tokens := ReadStream on: (self parseString: string) tokens.
	self
		assert: tokens next asAssociation equal: #'int16' -> 0;
		assert: tokens next asAssociation equal: #'int16' -> 16r7FFF;
		assert: tokens next asAssociation equal: #'uint16' -> 16rFFFF;
		assert: tokens next asAssociation equal: #'int32' -> 16r7FFFFFFF;
		assert: tokens next asAssociation equal: #'uint32' -> 16rFFFFFFFF;
		assert: tokens next asAssociation equal: #'int64' -> 16r7FFFFFFFFFFFFFFF;
		assert: tokens next asAssociation equal: #'uint64' -> 16rFFFFFFFFFFFFFFFF;

		assert: tokens next asAssociation equal: #'int64' -> 0;
		assert: tokens next asAssociation equal: #'int64' -> 0;
		assert: tokens next asAssociation equal: #'uint64' -> 1;
		assert: tokens next asAssociation equal: #'uint64' -> 1;
		assert: tokens next asAssociation equal: #'uint16' -> 2;
		assert: tokens next asAssociation equal: #'uint16' -> 2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testDefinedOperator

	| string tokens |
	string := '\
#define JAMES _a1_
#if defined(JAMES)
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testDefinedOperator2

	| string tokens |
	string := '\
#define JAMES _a1_
#if defined JAMES 
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testDefinedOperator3

	| string tokens |
	string := '\
#define JAMES
#if defined JAMES && !(defined BEVERLY || defined LAURA || defined JOHN || defined DANDY)
aaa
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testDefinedOperator4

	| string tokens |
	string := '
#define FOO(a,b) a + b
#define BAR(a) FOO((a),0) * 2
BAR(x)	//		(x) + 0 * 2
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 7;
		assert: (tokens at: 1) asAssociation equal: #'punctuator' -> '(';
		assert: (tokens at: 2) asAssociation equal: #'identifier' -> 'x';
		assert: (tokens at: 3) asAssociation equal: #'punctuator' -> ')';
		assert: (tokens at: 4) asAssociation equal: #'punctuator' -> '+';
		assert: (tokens at: 5) asAssociation equal: #'int16' -> 0;
		assert: (tokens at: 6) asAssociation equal: #'punctuator' -> '*';
		assert: (tokens at: 7) asAssociation equal: #'int16' -> 2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testDefinedOperator5

	| string tokens |
	string := '\
#if !defined _SIGNAL_H && !defined __need_siginfo_t \
    && !defined __need_sigevent_t
aaa
#endif
# define _SIGNAL_H
#if !defined _SIGNAL_H && !defined __need_siginfo_t \
    && !defined __need_sigevent_t
# error "Never include this file directly.  Use <signal.h> instead"
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 2;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'aaa';
		assert: (tokens at: 2) asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testDefinedOperator6

	| string tokens |
	string := '
#define FOO(a) a + a
FOO(a)
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 3;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'a';
		assert: (tokens at: 2) asAssociation equal: #'punctuator' -> '+';
		assert: (tokens at: 3) asAssociation equal: #'identifier' -> 'a';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testDirectiveSpacing

	| string tokens |
	string := '
 # ifdef MACRO
aaa
#	else
bbb
	#		endif /* MACRO */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testEmptyLine

	| string tokens |
	string := '
'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 0;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testEquals

	| string tokens |
	string := '
#if 1 == 1
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testFunctionPointerAsArgumentToFunction

	| string tokens token |
	string := '
extern int foo ( int ( * bar ) ( unsigned int ) ) ;
'.
	tokens := (self parseString: string) tokens.
	self
		assert: (token := tokens at: 3) isIdentifierToken;
		assert: token value equal: 'foo(';
		assert: (token := tokens at: 4) isIdentifierToken;
		assert: token value equal: 'int';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testGreedyToken
		" a ++ ++ + b "

	| string tokens |
	string := 'a+++++b'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 5;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'a';
		assert: (tokens at: 2) asAssociation equal: #'punctuator' -> '++';
		assert: (tokens at: 3) asAssociation equal: #'punctuator' -> '++';
		assert: (tokens at: 4) asAssociation equal: #'punctuator' -> '+';
		assert: (tokens at: 5) asAssociation equal: #'identifier' -> 'b';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testHexadecimalConstants

	| string tokens |
	string := '
0x0 0x7FFF 0xFFFF 0x7FFFFFFF 0xFFFFFFFF 0x7FFFFFFFFFFFFFFF 0xFFFFFFFFFFFFFFFF
0x0l 0x0L 0x1UL 0x1LU 0x2U 0x2u
'.
	tokens := ReadStream on: (self parseString: string) tokens.
	self
		assert: tokens next asAssociation equal: #'int16' -> 0;
		assert: tokens next asAssociation equal: #'int16' -> 16r7FFF;
		assert: tokens next asAssociation equal: #'uint16' -> 16rFFFF;
		assert: tokens next asAssociation equal: #'int32' -> 16r7FFFFFFF;
		assert: tokens next asAssociation equal: #'uint32' -> 16rFFFFFFFF;
		assert: tokens next asAssociation equal: #'int64' -> 16r7FFFFFFFFFFFFFFF;
		assert: tokens next asAssociation equal: #'uint64' -> 16rFFFFFFFFFFFFFFFF;

		assert: tokens next asAssociation equal: #'int64' -> 0;
		assert: tokens next asAssociation equal: #'int64' -> 0;
		assert: tokens next asAssociation equal: #'uint64' -> 1;
		assert: tokens next asAssociation equal: #'uint64' -> 1;
		assert: tokens next asAssociation equal: #'uint16' -> 2;
		assert: tokens next asAssociation equal: #'uint16' -> 2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testHexCharacterToken

	| string tokens |
	string := '''\x0'' ''\x7f'''.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 2;
		assert: (tokens at: 1) asAssociation equal: #'uint8' -> 0;
		assert: (tokens at: 1) source equal: '''\x0'' ';
		assert: (tokens at: 2) asAssociation equal: #'uint8' -> 16r7F;
		assert: (tokens at: 2) source equal: '''\x7f''';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testHexNumberInExpression

	| string tokens |
	string := '
#if 0x1F == 31
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testIfMacroFunction

	| string tokens |
	string := '
#define FOO(x) (x == 3)
#if 1 && 1 && FOO(3) && 1
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testIgnore

	| string tokens |
	string := '\
#if JAMES
#include missing
#else
bbb
#endif /* JAMES */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testIllegalStringLiteral

	| string |
	string := 'aaa"/* xx
xx */"bbb'.
	self
		should: [self parseString: string]
		raise: Error.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testIncompleteCommentInDefine

	| string tokens |
	string := '
#define FOO 1 /* part of comment 
		rest of comment */
FOO
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'int16' -> 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testLineComment

	| string tokens token |
	string := 'aaa//abcde
bbb'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 2;
		assert: (token := tokens at: 1) notNil;
		assert: token line equal: 1;
		assert: token isIdentifierToken;
		assert: token value equal: 'aaa';
		assert: token source equal: 'aaa//abcde';

		assert: (token := tokens at: 2) notNil;
		assert: token line equal: 2;
		assert: token isIdentifierToken;
		assert: token value equal: 'bbb';
		assert: token source first equal: Character lf;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testLines

	| string tokens token |
	string := 'aaa
bbb'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 2;
		assert: (token := tokens at: 1) notNil;
		assert: token line equal: 1;
		assert: token isIdentifierToken;
		assert: token value equal: 'aaa';
		assert: token source size equal: 3;

		assert: (token := tokens at: 2) notNil;
		assert: token line equal: 2;
		assert: token isIdentifierToken;
		assert: token value equal: 'bbb';
		assert: token source size equal: 4;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testLineSplit

	| string tokens |
	string := '/\
*
*/ # /*
*/ defi\
ne FO\
O 10\
20
FOO'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size equal: 1;
		assert: (tokens at: 1) asAssociation equal: #'int16' -> 1020;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testLongNumberInExpression

	| string tokens |
	string := '
#if __cplusplus >= 199707L 
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testMergeLines

	| string tokens token |
	string := 'aaa \
bb\
b'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 2;
		assert: (token := tokens at: 1) notNil;
		assert: token line equal: 1;
		assert: token isIdentifierToken;
		assert: token value equal: 'aaa';
		assert: token source size equal: 4;

		assert: (token := tokens at: 2) notNil;
		assert: token line equal: 3;
		assert: token isIdentifierToken;
		assert: token value equal: 'bbb';
		assert: token source size equal: 7;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testMultiLineBlockComment

	| string tokens token |
	string := 'aaa/* xxx 
yyy
zzz*/bbb'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 2;
		assert: (token := tokens at: 1) notNil;
		assert: token line equal: 3;
		assert: token isIdentifierToken;
		assert: token value equal: 'aaa';
		assert: token source size equal: 20;

		assert: (token := tokens at: 2) notNil;
		assert: token line equal: 3;
		assert: token isIdentifierToken;
		assert: token value equal: 'bbb';
		assert: token source size equal: 3;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testMultipleOperators

	| string tokens |
	string := '
#define FALSE 0
#define TRUE 1
#if FALSE || FALSE || TRUE
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testMultipleOperatorsA

	| string tokens |
	string := '
#if 1 == 1 && 2 == 2
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testMultipleOperatorsB

	| string tokens |
	string := '
#if 2 + 3 * 4 == 14
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testMultipleOperatorsC

	| string tokens |
	string := '
#if 2 + 4 / 2 * 3 - 4 == 4
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testNestedConditionals

	| string tokens |
	string := '
#if 0
#if 1
aaa
#else
bbb
#endif
#if 1
xxx
#endif
#else
ccc
#endif
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'ccc';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testNonFunctionMacro
	"http://gcc.gnu.org/onlinedocs/gcc-3.0.2/cpp_3.html
	() c_init()()
	"

	| string tokens |
	string := '
#define lang_init ()    c_init()
lang_init()'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 6;
		assert: (tokens at: 1) asAssociation equal: #'punctuator' -> '(';
		assert: (tokens at: 2) asAssociation equal: #'punctuator' -> ')';
		assert: (tokens at: 3) asAssociation equal: #'identifier' -> 'c_init(';
		assert: (tokens at: 4) asAssociation equal: #'punctuator' -> ')';
		assert: (tokens at: 5) asAssociation equal: #'punctuator' -> '(';
		assert: (tokens at: 6) asAssociation equal: #'punctuator' -> ')';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testNotDefinedOperator

	| string tokens |
	string := '\
#define JAMES _a1_
#if !defined(JAMES)
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testNumber

	| string tokens |
	string := '
#if 1
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testOctalCharacterToken

	| string tokens |
	string := '''\0'' ''\01'' ''\011'' ''\0111'''.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 4;
		assert: (tokens at: 1) asAssociation equal: #'uint8' -> 0;
		assert: (tokens at: 2) asAssociation equal: #'uint8' -> 1;
		assert: (tokens at: 3) asAssociation equal: #'uint8' -> 9;
		assert: (tokens at: 4) asAssociation equal: #'uint8' -> 73;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testOctalConstants

	| string tokens |
	string := '
0 077777 0177777 017777777777 037777777777 0777777777777777777777 01777777777777777777777
0l 0L 01UL 01LU 02U 02u
'.
	tokens := ReadStream on: (self parseString: string) tokens.
	self
		assert: tokens next asAssociation equal: #'int16' -> 0;
		assert: tokens next asAssociation equal: #'int16' -> 16r7FFF;
		assert: tokens next asAssociation equal: #'uint16' -> 16rFFFF;
		assert: tokens next asAssociation equal: #'int32' -> 16r7FFFFFFF;
		assert: tokens next asAssociation equal: #'uint32' -> 16rFFFFFFFF;
		assert: tokens next asAssociation equal: #'int64' -> 16r7FFFFFFFFFFFFFFF;
		assert: tokens next asAssociation equal: #'uint64' -> 16rFFFFFFFFFFFFFFFF;

		assert: tokens next asAssociation equal: #'int64' -> 0;
		assert: tokens next asAssociation equal: #'int64' -> 0;
		assert: tokens next asAssociation equal: #'uint64' -> 1;
		assert: tokens next asAssociation equal: #'uint64' -> 1;
		assert: tokens next asAssociation equal: #'uint16' -> 2;
		assert: tokens next asAssociation equal: #'uint16' -> 2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testOrOperator

	| string tokens |
	string := '\
#define JAMES _a1_
#if defined(JAMES) || defined(BEVERLY)
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testOrOperator2

	| string tokens |
	string := '\
#if defined(JAMES) || defined(BEVERLY)
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testParenthesis

	| string tokens |
	string := '(x)'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 3;
		assert: (tokens at: 1) asAssociation equal: #'punctuator' -> '(';
		assert: (tokens at: 2) asAssociation equal: #'identifier' -> 'x';
		assert: (tokens at: 3) asAssociation equal: #'punctuator' -> ')';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testParenthesis2

	| string tokens |
	string := '
#if defined(James) || (defined(Beverly))
aaa
#else
bbb
#endif'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testParenthesisProvidedByMacro

	| string tokens token |
	string := '
# define __MALLOC_P(args)	args __THROW
# define __attribute_malloc__ __attribute__ ((__malloc__))
# define __wur /* Ignore */
/* Allocate SIZE bytes of memory.  */
extern void *malloc __MALLOC_P ((size_t __size)) __attribute_malloc__ __wur;
'.
	tokens := (self parseString: string) tokens.
	self
		assert: 4 < tokens size;
		assert: (token := tokens at: 4) asAssociation equal: #'identifier' -> 'malloc(';
		assert: (token := tokens at: 9) asAssociation equal: #'identifier' -> '__attribute__(';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testPunctuatorToken

	| string tokens |
	string := '2 * 3'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 3;
		assert: (tokens at: 1) asAssociation equal: #'int16' -> 2;
		assert: (tokens at: 2) asAssociation equal: #'punctuator' -> '*';
		assert: (tokens at: 3) asAssociation equal: #'int16' -> 3;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testRedefine
"cpp -E << EOF
#define AAA 1
#define BBB AAA
BBB
#define AAA 2
BBB
EOF
"
	| string preprocessor tokens definitions |
	string := '
#define AAA 1
#define BBB AAA
BBB
#define AAA 2
BBB
'.
	preprocessor := self parseString: string.
	tokens := preprocessor tokens.
	definitions := preprocessor definitions.
	self
		assert: tokens size equal: 2;
		assert: (tokens at: 1) asAssociation equal: #'int16' -> 1;
		assert: (tokens at: 2) asAssociation equal: #'int16' -> 2;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testStringToken

	| string preprocessor tokens |
	string := 'printf "ab\ncd\ref\tgh\''ij\"kl\?mn\\op\0" "a\''c"'.
	preprocessor := self parseString: string.
	self 
		assert: (tokens := preprocessor tokens) size equal: 3;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'printf';
		assert: (tokens at: 2) key equal: #'string';
		assert: (string := (tokens at: 2) value) size equal: 24;
		assert: (string copyFrom: 1 to: 2) equal: 'ab';
		assert: (string at: 3) equal: Character cr;
		assert: (string copyFrom: 4 to: 5) equal: 'cd';
		assert: (string at: 6) equal: Character lf;
		assert: (string copyFrom: 7 to: 8) equal: 'ef';
		assert: (string at: 9) equal: Character tab;
		assert: (string copyFrom: 10 to: 11) equal: 'gh';
		assert: (string at: 12) equal: $';
		assert: (string copyFrom: 13 to: 14) equal: 'ij';
		assert: (string at: 15) equal: $";
		assert: (string copyFrom: 16 to: 17) equal: 'kl';
		assert: (string at: 18) equal: $?;
		assert: (string copyFrom: 19 to: 20) equal: 'mn';
		assert: (string at: 21) equal: $\;
		assert: (string copyFrom: 22 to: 23) equal: 'op';
		assert: (string at: 24) codePoint equal: 0;
		assert: (tokens at: 3) asAssociation equal: #'string' -> 'a''c';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testTwoComments

	| string tokens |
	string := 'aaa/* x */bbb/* y */ccc'.
	tokens := (self parseString: string) tokens.
	self 
		assert: tokens size = 3;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'aaa';
		assert: (tokens at: 1) source equal: 'aaa/* x */';
		assert: (tokens at: 2) asAssociation equal: #'identifier' -> 'bbb';
		assert: (tokens at: 3) asAssociation equal: #'identifier' -> 'ccc';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
testWideCharacterToken

	| string preprocessor tokens |
	string := '
#if L''\0'' - 1 > 0
aaa
#else
bbb
#endif
'.
	preprocessor := self parseString: string.
	self 
		assert: (tokens := preprocessor tokens) size equal: 1;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define

	| string preprocessor tokens definitions definition |
	string := '#define JAMES _a1_'.
	preprocessor := self parseString: string.
	tokens := preprocessor tokens.
	definitions := preprocessor definitions.
	self
		assert: tokens isEmpty;
		assert: (definitions includesKey: 'JAMES');
		assert: ((definition := definitions at: 'JAMES') isKindOf: CPreprocessorToken);
		assert: definition defineString equal: '_a1_';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define_cascade

	| string tokens |
	string := '
#define b a
#define c b
c
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'a';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define_commaInParameter

	| string tokens |
	string := '
#define foo(x) x + 1
foo ((1,2))
'.	
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 7;
		assert: (tokens at: 1) asAssociation equal: #'punctuator' -> '(';
		assert: (tokens at: 2) asAssociation equal: #'int16' -> 1;
		assert: (tokens at: 3) asAssociation equal: #'punctuator' -> ',';
		assert: (tokens at: 4) asAssociation equal: #'int16' -> 2;
		assert: (tokens at: 5) asAssociation equal: #'punctuator' -> ')';
		assert: (tokens at: 6) asAssociation equal: #'punctuator' -> '+';
		assert: (tokens at: 7) asAssociation equal: #'int16' -> 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define_empty

	| string preprocessor tokens definitions definition |
	string := '#define JAMES'.
	preprocessor := self parseString: string.
	tokens := preprocessor tokens.
	definitions := preprocessor definitions.
	self
		assert: tokens isEmpty;
		assert: (definitions includesKey: 'JAMES');
		assert: ((definition := definitions at: 'JAMES') isKindOf: CPreprocessorToken);
		assert: definition defineString equal: '';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define_expression

	| string tokens |
	string := '
#define PLUS(x, y) ((x) + (y))
PLUS(1,2)'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 9;
		assert: (tokens at: 1) asAssociation equal: #'punctuator' -> '(';
		assert: (tokens at: 1) source equal: Character lf asString , 'PLUS(1,2)';
		assert: (tokens at: 2) asAssociation equal: #'punctuator' -> '(';
		assert: (tokens at: 2) source equal: '';
		assert: (tokens at: 3) asAssociation equal: #'int16' -> 1;
		assert: (tokens at: 4) asAssociation equal: #'punctuator' -> ')';
		assert: (tokens at: 5) asAssociation equal: #'punctuator' -> '+';
		assert: (tokens at: 6) asAssociation equal: #'punctuator' -> '(';
		assert: (tokens at: 7) asAssociation equal: #'int16' -> 2;
		assert: (tokens at: 8) asAssociation equal: #'punctuator' -> ')';
		assert: (tokens at: 9) asAssociation equal: #'punctuator' -> ')';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define_functionSpace

	| string tokens |
	string := '
#define foo(a) a + 1
foo (2)'.	
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 3;
		assert: (tokens at: 1) asAssociation equal: #'int16' -> 2; 
		assert: (tokens at: 2) asAssociation equal: #'punctuator' -> '+';
		assert: (tokens at: 3) asAssociation equal: #'int16' -> 1;
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define_multipleTokens

	| string preprocessor tokens definitions definition |
	string := '#define JAMES int *'.
	preprocessor := self parseString: string.
	tokens := preprocessor tokens.
	definitions := preprocessor definitions.
	self
		assert: tokens isEmpty;
		assert: (definitions includesKey: 'JAMES');
		assert: (definition := (definitions at: 'JAMES')) notNil;
		assert: definition defineString equal: 'int *';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define_recursive

	| string tokens |
	string := '
#define a a
a
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: (tokens at: 1) asAssociation equal: #'identifier' -> 'a';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_define_withMacro

	| string preprocessor tokens definitions definition |
	string := '
#define TRUE 1
#define JAMES TRUE
'.
	preprocessor := self parseString: string.
	tokens := preprocessor tokens.
	definitions := preprocessor definitions.
	self
		assert: tokens isEmpty;
		assert: (definitions includesKey: 'JAMES');
		assert: ((definition := definitions at: 'JAMES') isKindOf: CPreprocessorToken);
		assert: definition defineString equal: 'TRUE';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_elif

	| string tokens x |
	string := '
#define JAMES 1
#if JAMES == 1
aaa
#elif JAMES == 2
bbb
#else
#error "Should not get here!"
#endif /* JAMES */
'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_elif_2

	| string tokens x |
	string := '
#define JAMES 3
#if JAMES == 1
aaa
#elif JAMES == 2
bbb
#elif JAMES == 3
ccc
#else
#error "Should not get here!"
#endif /* JAMES */
'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'ccc'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_if

	| string tokens x |
	string := '
#define JAMES 0
#if JAMES
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_if2

	| string tokens x |
	string := '
#if JAMES
aaa
#else
bbb
#endif /* JAMES */
'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_if3

	| string tokens x |
	string := '
#define JAMES
#if JAMES
aaa
#else
bbb
#endif /* JAMES */
'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_if4

	| string tokens x |
	string := '
#if 0
#   if SHLIB_COMPAT (libc, GLIBC_2_0, GLIBC_2_1)
#   endif
#else
abc
#endif'.	
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'abc'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_if5
"#if ((!defined __STRICT_ANSI__ || (_XOPEN_SOURCE - 0) >= 500) && \
     !defined _POSIX_SOURCE && !defined _POSIX_C_SOURCE)"

	| string tokens x |
	string := '
#define JAMES 0
#if \
 !JAMES)
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_ifdef

	| string tokens x |
	string := '
#define JAMES 0
#ifdef JAMES
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_ifdef2

	| string tokens x |
	string := '
#ifdef JAMES
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_ifdef_defined

	| string tokens |
	string := '\
#define JAMES _a1_
#ifdef JAMES
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_ifdef_undefined

	| string tokens |
	string := '
#ifdef MACRO
aaa
#else
bbb
#endif /* MACRO */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_ifndef

	| string tokens x |
	string := '
#ifndef JAMES
	#define JAMES 1
	aaa
	#ifndef JAMES
		bbb
	#endif
	ccc
#endif'.
	tokens := (x := self parseString: string) tokens.
	self
		assert: tokens size equal: 2;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa'; 
		assert: tokens last asAssociation equal: #'identifier' -> 'ccc'; 
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_ifndef_undefined

	| string tokens |
	string := '
#ifndef MACRO
aaa
#else
bbb
#endif /* MACRO */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_if_FALSE

	| string tokens |
	string := '
#define JAMES FALSE
#if JAMES
aaa
#else
bbb
#endif /* JAMES */
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'bbb';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_if_TRUE

	| string tokens |
	string := '
#define TRUE 1
#define JAMES TRUE
#if JAMES
aaa
#else
bbb
#endif /* JAMES */'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 1;
		assert: tokens first asAssociation equal: #'identifier' -> 'aaa';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_undef

	| string tokens token |
	string := '
#define JAMES 1
JAMES
#undef JAMES
JAMES
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 2;
		assert: (token := tokens at: 1) notNil;
		assert: token type equal: #'int16';
		assert: token value equal: 1;
		assert: token line equal: 3;
		assert: token source equal: Character lf asString , 'JAMES';

		assert: (token := tokens at: 2) notNil;
		assert: token isIdentifierToken;
		assert: token value equal: 'JAMES';
		assert: token line equal: 5;
		assert: token source equal: Character lf asString , 'JAMES';
		yourself.
%
category: 'passing'
set compile_env: 0
method: CPreprocessorTestCase
test_undef_missing

	| string tokens |
	string := '
#define JAMES 1
JAMES
#undef JAMES
#undef JAMES
JAMES
'.
	tokens := (self parseString: string) tokens.
	self
		assert: tokens size equal: 2;
		assert: (tokens at: 2) asAssociation equal: #'identifier' -> 'JAMES';
		yourself.
%
