/*
	webtools.js
*/

GemStone = function() {	//	hide everything inside an anonymous function to isolate namespace
	var requestCount = 0	//	keep a cumulative count of round-trips to server
	,	tools = {}
	,	$tabs
	,	$statusBar
	,	nextId = 0
	;
	$(document).ready(function() {
		$tabs = $('#tabs');
		$statusBar = $('#statusBar');
		createMainTabs();
		getJSON('tools', null, gotTools);
		$(window).resize(resize);	//	register to get resize events
		return;
		
		function createMainTabs() {
			$tabs.tabs({ show: showTab });
			$('.ui-tabs-nav', $tabs)
				.removeClass('ui-corner-all')
				.addClass('ui-corner-top');
			return;
			
			function showTab(event, ui) { 
				var $tab = $('#' + ui.panel.id)
				,	tool = $tab.data('tool')
				;
				$(window).resize();
				if ( tool && tool.onShow ) {
					tool.onShow($tab[0]);
				}
			}
		}

		function gotTools(json) { 
			var items = [];
			$.each(json.tools, function(){
				items.push('<tr class="clickable" title="' + this.file + '">');
				items.push('<td>' + this.name + '</td>');
				items.push('<td>' + this.description + '</td>');
				items.push('</tr>');
			});
			$('#homePanel tbody').empty().append(items.join(''));
			$('#homePanel tbody tr').click(onClick);
			return;

			function onClick(event) {
				event.preventDefault();
				$.ajax({ 
					url: $(this).attr('title'), 
					success: function(html) { $('body').append(html); } 
				});
			}
		}
	});

	return {
		addTab:		addTab
	,	getJSON:	getJSON
	,	nextId:		getNextId
	,	scroll:		scroll
	}

	function addTab(tool) {
		var selector = '#' + getNextId();
		$('#' + tool.id).attr('id', selector.substring(1));
		moveOnlyOnce($(selector));
		$tabs.tabs('add', selector, tool.label);
		var $listItem = $('> ul > li:last', $tabs);
		$listItem.attr('title', tool.title);
		addCloseButton($listItem);
		$(selector).data('tool', tool);
		$tabs.tabs('select', selector);
		if (tool.onAdd) { tool.onAdd($(selector)); }
		return;

		function addCloseButton($listItem) {
			$listItem.append('<a href="#" class="closeTab">X</a>');
			$('a.closeTab', $listItem).click(removeTab);
			return;
			
			function removeTab(event) {
				var list = $listItem.parent().children('li')
				,	index = list.index($listItem)
				;
				$(selector).remove();
				$tabs.tabs('remove', index);
				event.preventDefault();
			}
		};

		function moveOnlyOnce($tabPanel) {
			var	$div = $('.once', $tabPanel)
			,	newClass = $div.attr('id')
			;
			$div.addClass(newClass);
			if (1 === $('.' + newClass).length) {
				$div.attr('id', null);
				$('style', $div).appendTo('head');
				$('style', $div).remove();
				if ($div.html().replace(/(^\s+)(\s+$)/, "").length) {
					$('body').append($div);
				}
			} else {
				$div.remove();
			}
		}
	}

	function getJSON(url, arguments, callback) { 
		var startTime = new Date().getTime();
		$statusBar.text('Sent request #' + (++requestCount) + ' for ' + url);
		$.getJSON(url, arguments, success);
		return;
		
		function success(json) {
			var serverTime = json._time;
			var networkTime = new Date().getTime() - startTime - serverTime;
			startTime = new Date().getTime();
			var error = json._error;
			if (error) {
				alert(error + ' (see console log for stack)');
				console.log(json._stack);
			} else {
				(callback)(json);
			};
			var elapsed = new Date().getTime() - startTime;
			$statusBar.text(
				'Request for ' + url + ' (roundtrip #' + requestCount + ') took ' +
				serverTime + ' ms on server, ' + networkTime + ' ms on the network, and ' 
				+ elapsed + ' ms on the client.'
			);
		};
	}

	function getNextId() {
		nextId = nextId + 1;
		return 'id' + nextId;
	}

	function scroll($div) {
		if ( 1 === $('table', $div).length ) { 
			var table = $('table', $div).html();
			$div.empty().append(
				'<div class="tableHeader"><table>' + table + '</table></div>' +
				'<div class="tableBody"><table>' + table + '</table></div>'
				);
			$('.tableHeader tbody', $div).remove();
			$('.tableBody caption', $div).remove();
			$div.addClass('scrollingTable');
		}
	}
	
	function resize() {
		var $tables = $('.ui-tabs-panel .scrollingTable')
						.not('.ui-tabs-hide .scrollingTable');
		$.each($tables, function() { 
			setColumnWidths($('.tableHeader table', this), $('.tableBody table', this));
		});
		return;

		function setColumnWidths($header, $body) {
			if (0 === $('thead tr th.filler', $header).length) {
				$('thead tr', $header).append('<th class="filler"></th>');
			}
			var emptyBody = 0 === $('tbody tr', $body).length
			,	columnCount = $('thead tr th', $body).length - 1
			;
			$('thead', $body).removeClass('hide');
			clearWidths($header);
			$('thead tr th.filler', $header).css('width', '15px');
			clearWidths($body);
			if (emptyBody) {
				$('tbody', $body).append('<tr />');
				for (var i = 0; i < columnCount; ++i) {
					$('tbody tr', $body).append('<td>&nbsp</td>');
				}
			}
			var widths = $.map($('tbody tr:first td', $body), calculateWidth);
			setWidths($body);
			widths[0] = widths[0] + 1;
			widths.push(15);
			setWidths($header);
			if (emptyBody) {
				$('tbody', $body).empty();
			}
			$('thead', $body).addClass('hide');
			$body.parent().css('top', $header.outerHeight() - 3 + 'px');
			$body.
			return;
			
			function clearWidths($element) {
				$('thead tr:first th', $element).each(function() {
					$(this).removeAttr('style');
				});
				$element.removeAttr('style');
			}
			function setWidths($element) {
				var total = 0;
				$('thead tr:first th', $element).each(function(index) {
					$(this).css({ width: widths[index] + 'px' });
					total = total + widths[index];
				});
				$element.css({'table-layout': 'fixed', 'width': total});
			}
			function calculateWidth(td) {	
				if ($.browser.msie) 	{ return $(td).outerWidth() - 1; 	}
				if ($.browser.mozilla) 	{ return $(td).width();				}
				return $(td).outerWidth();
			}
		}	
	}
	
}();
