! ========================================================================
! Copyright (C) by VMware, Inc. 1991-2011.  All Rights Reserved
! ========================================================================

expectvalue %String
run
(Object 
	subclass: 'CHeader'
	instVarNames: #( stream types structs
                    unions functions enums storage
                    declarations preprocessor source)
	classVars: #()
	classInstVars: #()
	poolDictionaries: { }
	inDictionary: Globals
	options: #( #logCreation )) definition
%

! set class category
run
CHeader category: 'FFI'.
true
%

! Remove existing behavior from CHeader
doit
CHeader removeAllMethods.
CHeader class removeAllMethods.
true
%
! ------------------- Class methods for CHeader
category: 'Documentation'
set compile_env: 0
classmethod: CHeader
comment
^'This class is part of the FFI implementation and represents a C header file. 
It is typically used to generate a Smalltalk class that wraps a function or structure.
Usage examples can be found in the following instance methods:
	#wrapperForLibraryAt:
	#wrapperNamed:forLibraryAt:select:
	#wrapperForTypeNamed:
	#wrapperNamed:forStruct:
'
%
category: 'other'
set compile_env: 0
classmethod: CHeader
path: aString

	^self preprocessor: (CPreprocessor parseFileAt: aString).
%
category: 'other'
set compile_env: 0
classmethod: CHeader
preprocessor: aCPreprocessor

	^self new
		initialize: aCPreprocessor;
		yourself.
%
category: 'other'
set compile_env: 0
classmethod: CHeader
string: aString

	^self preprocessor: (CPreprocessor parseString: aString).
%
! ------------------- Instance methods for CHeader
category: 'CLibrary'
set compile_env: 0
method: CHeader
add: aCDeclaration for: aCLibrary to: aClass

	aClass addClassVarName: aCDeclaration asClassVarName.
	aClass class
		compileMethod: aCDeclaration createFunctionInitializerCode
		dictionaries: System myUserProfile symbolList
		category: 'Functions' environmentId: 0 .
	aClass
		compileMethod: aCDeclaration invokeFunctionCode
		dictionaries: SymbolList new
		category: 'Functions' environmentId: 0.
%
category: 'Accessing'
set compile_env: 0
method: CHeader
atEnd

	^stream atEnd.
%
category: 'CLibrary'
set compile_env: 0
method: CHeader
createDefinesInClass: aClass
	"preprocessor createDefinesInClass: aClass"
%
category: 'CLibrary'
set compile_env: 0
method: CHeader
createFunctionsInClass: newClass library: library select: aBlock

	| localFunctions |
	localFunctions := aBlock isNil 
		ifTrue: [functions]
		ifFalse: [functions select: aBlock].
	localFunctions do: [:each | 
		self 
			add: each 
			for: library
			to: newClass.
	].
	self 
		createInitializerInClass: newClass
		forFunctions: localFunctions 
		library: library.
%
category: 'CLibrary'
set compile_env: 0
method: CHeader
createInitializerInClass: aClass forFunctions: someFunctions library: library

	aClass class
		compileMethod: (self initializeFunctionsCodeForFunctions: someFunctions library: library)
		dictionaries: System myUserProfile symbolList
		category: 'Functions' environmentId: 0 .
%
category: 'Accessing'
set compile_env: 0
method: CHeader
declarations

   ^declarations
%
category: 'Accessing'
set compile_env: 0
method: CHeader
enums

   ^enums
%
category: 'Accessing'
set compile_env: 0
method: CHeader
fieldsForStruct: aString

   ^(structs at: aString) fields.
%
category: 'Accessing'
set compile_env: 0
method: CHeader
functions

   ^functions
%
category: 'other'
set compile_env: 0
method: CHeader
initialize: aCPreprocessor
 
	preprocessor := aCPreprocessor.
	stream := ReadStream on: preprocessor tokens.
	types := KeyValueDictionary new.
	structs := KeyValueDictionary new.
	unions := KeyValueDictionary new.
	functions := KeyValueDictionary new.
	enums := KeyValueDictionary new.
	storage := KeyValueDictionary new.
	declarations := OrderedCollection new.

	[
		self skipEmpty.
		stream atEnd not.
	] whileTrue: [
		source := String new.
		self readOne.
	].
%
category: 'CLibrary'
set compile_env: 0
method: CHeader
initializeFunctionsCodeForFunctions: someFunctions library: library

	| myStream |
	myStream := WriteStream on: String new.
	myStream 
		nextPutAll: 'initializeFunctions

	| library |
	library := CLibrary named: ''';
		nextPutAll: library name;
		nextPutAll: '''.'; lf; tab;
		nextPutAll: 'self'; lf; tab; tab;
		yourself.
	someFunctions do: [:aFunction |
		myStream 
			nextPutAll: aFunction initializerFunctionName;
			nextPutAll: ' library;'; lf; tab; tab;
			yourself.
	].
	^myStream 
		nextPutAll: 'yourself.'; lf;
		contents.
%
category: 'CLibrary'
set compile_env: 0
method: CHeader
newClassNamed: aString

	^self
		newClassNamed: aString
		instVarNames: #().
%
category: 'CLibrary'
set compile_env: 0
method: CHeader
newClassNamed: aString instVarNames: anArray

	^Object
		subclass: aString
		instVarNames: anArray
		classVars: #()
		classInstVars: #()
		poolDictionaries: #()
		inDictionary: SymbolDictionary new "GsPackagePolicy current homeSymbolDict"
		options: #().
%
category: 'Accessing'
set compile_env: 0
method: CHeader
next

	| token |
	token := stream next.
	source notNil ifTrue: [source add: token source].
	^token.
%
category: 'Accessing'
set compile_env: 0
method: CHeader
peek

	^stream peek.
%
category: 'other'
set compile_env: 0
method: CHeader
readClass

	| token classDeclaration oldTypes oldStructs oldUnions oldFunctions oldEnums oldStorage newStorage |
	classDeclaration := declarations last.
	(token := self peek) isSemicolonToken ifTrue: [
		self next. 
		classDeclaration source add: source.
		source := nil.
		^self.
	].
	types 
		at: classDeclaration name 
		put: classDeclaration.
	oldTypes := types copy.
	oldStructs := structs copy.
	oldUnions := unions copy.
	oldFunctions := functions copy.
	oldEnums := enums copy.
	oldStorage := storage copy.
	[
		self skipEmpty.
		(token := self peek) isCloseCurlyBracketToken not.
	] whileTrue: [
		(token type == #'identifier' and: [token value = 'public' or: [token value = 'private']]) ifTrue: [
			(token := self next; next) isColonToken ifFalse: [self error: 'Unexpected token'].
		].
		self readOne.
	].
	self next.
	types := oldTypes.
	structs := oldStructs.
	unions := oldUnions.
	functions := oldFunctions.
	enums := oldEnums.
	newStorage := storage.
	storage := oldStorage.
	storage keys do: [:each | newStorage removeKey: each].
	newStorage := newStorage asArray collect: [:each | 
		(declarations indexOf: each) -> each.
	].
	newStorage := newStorage asSortedCollection asArray collect: [:each | each value].
	classDeclaration setClassProperties: newStorage.
	self atEnd ifFalse: [
		(token := self next) isSemicolonToken ifFalse: [self error: 'Unexpected token'].
	].
	classDeclaration source: source.
%
category: 'other'
set compile_env: 0
method: CHeader
readExternC

	[
		self skipEmpty.
		self atEnd not.
	] whileTrue: [
		| declaration |
		source := String new.
		self peek isCloseCurlyBracketToken ifTrue: [self next. ^self].
		self readOne.
		((declaration := declarations last) isFunction and: [declaration storage == #'auto']) ifTrue: [
			declaration beExternStorage.
		].
	].
	self error: 'Expected end to extern "C"'.
%
category: 'other'
set compile_env: 0
method: CHeader
readOne
 
	| declaration token |
	declaration := CDeclaration header: self.
	declaration source: source.
	declarations add: declaration.
	declaration includesCode ifTrue: [^self].
	(declaration storage == #'extern' and: [declaration name = 'C' and: [declaration type == nil]]) ifTrue: [^self readExternC].
	(declaration storage == #'typedef' and: [declaration type == #'class']) ifTrue: [^self readClass].

	[
		self saveDeclaration: declaration.
		self atEnd ifTrue: [^self].
		(token := self next) isSemicolonToken ifTrue: [^self].
		token isCommaToken ifFalse: [self error: 'Expected comma or semicolon!'].
		true.
	] whileTrue: [
		declaration := declaration copy.
		declaration readDeclarator.
		declarations add: declaration.
	].
%
category: 'other'
set compile_env: 0
method: CHeader
saveDeclaration: aDeclaration
 
	| dictionary |
	aDeclaration name isNil ifTrue: [^self].
	dictionary := 
		aDeclaration storage = #'typedef' ifTrue: [types] ifFalse: [
		(aDeclaration isFunction and: [aDeclaration isStorage not]) ifTrue: [functions] ifFalse: [
		storage]].
	dictionary
		at: aDeclaration name
		put: aDeclaration.
%
category: 'other'
set compile_env: 0
method: CHeader
skipEmpty

	[
		self atEnd not and: [self peek key == #'empty'].
	] whileTrue: [
		self next.
	].
%
category: 'Accessing'
set compile_env: 0
method: CHeader
storage

   ^storage
%
category: 'Accessing'
set compile_env: 0
method: CHeader
structs

   ^structs
%
category: 'Accessing'
set compile_env: 0
method: CHeader
types

   ^types
%
category: 'Accessing'
set compile_env: 0
method: CHeader
unions

   ^unions
%
category: 'CLibrary'
set compile_env: 0
method: CHeader
wrapperForLibraryAt: aString
"
| header class |
header := CHeader path: '$GEMSTONE/include/gci.hf'.
class := header wrapperForLibraryAt: '$GEMSTONE/lib/libgcilnk.so'.
UserGlobals at: class name put: class.
"
	| name |
	name := ((aString subStrings: $/) last subStrings: $.) first.
	(3 < name size and: [(name copyFrom: 1 to: 3) = 'lib']) ifTrue: [name := name copyFrom: 4 to: name size].
	name first isLowercase ifTrue: [name at: 1 put: name first asUppercase].
	^self
		wrapperNamed: name
		forLibraryAt: aString 
		select: nil.
%
category: 'CByteArray'
set compile_env: 0
method: CHeader
wrapperForType: aCDeclaration

	^self
		wrapperNamed: aCDeclaration name
		forType: aCDeclaration.
%
category: 'CByteArray'
set compile_env: 0
method: CHeader
wrapperForTypeNamed: aString
"
| header class |
UserGlobals removeKey: #'GciErrSType' ifAbsent: [].
header := CHeader path: '$GEMSTONE/include/gci.hf'.
class := header wrapperForTypeNamed: 'GciErrSType'.
UserGlobals at: class name put: class.
"
	^self
		wrapperNamed: aString
		forType: (types at: aString).
%
category: 'CLibrary'
set compile_env: 0
method: CHeader
wrapperNamed: nameString forLibraryAt: pathString select: aBlock
"
| header class |
UserGlobals removeKey: #'GciLibrary' ifAbsent: [].
header := CHeader path: '$GEMSTONE/include/gci.hf'.
class := header
	wrapperNamed: 'GciLibrary'
	forLibraryAt: '$GEMSTONE/lib/libgcilnk.so'
	select: [:each | each name first = $G].
class initializeFunctions.
UserGlobals at: class name put: class.
"
	| library newClass |
	library := CLibrary named: pathString.
	newClass := self newClassNamed: nameString.
	self 
		createFunctionsInClass: newClass library: library select: aBlock;
		createDefinesInClass: newClass.
	^newClass.
%
category: 'CByteArray'
set compile_env: 0
method: CHeader
wrapperNamed: nameString forStruct: aCDeclaration
"
| header class |
UserGlobals removeKey: #'GciErrSType' ifAbsent: [].
header := CHeader path: '$GEMSTONE/include/gci.hf'.
class := header
	wrapperNamed: 'GciErrSType'
	forType: 'GciErrSType'.
UserGlobals at: class name put: class.
"
	| class symbolList string writeStream |
	symbolList := System myUserProfile symbolList.
	class := self 
		newClassNamed: nameString
		instVarNames: #(cByteArray).
	(writeStream := WriteStream on: String new)
		nextPutAll: 'COMMENT'; lf;
		yourself.
	aCDeclaration addSourceTo: writeStream.
	class class
		compileMethod: writeStream contents  "may signal CompileError or CompileWarning"
		dictionaries: symbolList
		category: 'Documentation' environmentId: 0 .
	string := 'on: aCByteArray

	^self basicNew
		initialize: aCByteArray;
		yourself.
'.
	class class
		compileMethod: string  "may signal CompileError or CompileWarning"
		dictionaries: symbolList
		category: 'Instance Creation' environmentId: 0 .
	string := 'new

	^self basicNew
		initialize;
		yourself.
'.
	class class
		compileMethod: string  "may signal CompileError or CompileWarning"
		dictionaries: symbolList
		category: 'Instance Creation' environmentId: 0 .
	string := 'initialize

	cByteArray := CByteArray new: ' , aCDeclaration byteSize printString , '.
'.
	class
		compileMethod: string  "may signal CompileError or CompileWarning"
		dictionaries: symbolList
		category: 'Initialization' environmentId: 0 .
	string := 'initialize: aCByteArray

	cByteArray := aCByteArray.
'.
	class
		compileMethod: string  "may signal CompileError or CompileWarning"
		dictionaries: symbolList
		category: 'Initialization' environmentId: 0 .
	aCDeclaration sourceStringsForAccessors do: [:each | 
		class
			compileMethod: each  "may signal CompileError or CompileWarning"
			dictionaries: symbolList
			category: 'Accessing' environmentId: 0 .
	].
	aCDeclaration sourceStringsForUpdators do: [:each | 
		class
			compileMethod: each  "may signal CompileError or CompileWarning"
			dictionaries: symbolList
			category: 'Updating' environmentId: 0 .
	].
	^class.
%
category: 'CByteArray'
set compile_env: 0
method: CHeader
wrapperNamed: nameString forType: aCDeclaration

	| struct |
	aCDeclaration type == #'class' ifTrue: [
		^self
			wrapperNamed: nameString
			forStruct: aCDeclaration.
	].
	(((struct := aCDeclaration type) isKindOf: CDeclaration) and: [struct type == #'struct']) ifFalse: [self error: 'Should be a struct'].
	^self
		wrapperNamed: nameString
		forStruct: struct.
%
