! ========================================================================
! Copyright (C) by VMware, Inc. 1991-2011.  All Rights Reserved
! ========================================================================

expectvalue %String
run
(Object
	subclass: 'CPreprocessor'
	instVarNames: #( path line tokens
                    readStream definitions searchPaths header)
	classVars: #( DefaultDefinitions)
	classInstVars: #()
	poolDictionaries: { }
	inDictionary: Globals
	options: #( #logCreation ))
definition
%

! set class category
run
CPreprocessor category: 'FFI'.
true
%

! Remove existing behavior from CPreprocessor
doit
CPreprocessor removeAllMethods.
CPreprocessor class removeAllMethods.
true
%
! ------------------- Class methods for CPreprocessor
category: 'Documentation'
set compile_env: 0
classmethod: CPreprocessor
comment
^'This class is internal to the FFI implementation and 
implements a preprocessor for a C header file.'
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessor
clearDefaultDefinitions
"
	CPreprocessor clearDefaultDefinitions.
"
	DefaultDefinitions := nil.
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessor
defaultDefinitions
"
	CPreprocessor defaultDefinitions 
		at: 'x86-64 (Linux)'.
"
	DefaultDefinitions isNil ifTrue: [
		DefaultDefinitions := KeyValueDictionary new.
		DefaultDefinitions objectSecurityPolicy: nil.
	].
	^DefaultDefinitions.
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessor
evaluateHeader: aCHeader

	^self basicNew evaluateHeader: aCHeader.
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessor
new

	^self basicNew
		initialize;
		yourself.
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessor
parseFileAt: aString

	^self new
		includePath: aString;
		yourself.
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessor
parseString: aString

	^self new
		parseString: aString;
		yourself.
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessor
references

^'
	http://www.acm.uiuc.edu/webmonkeys/book/c_guide/index.html
	http://gcc.gnu.org/onlinedocs/gcc-3.0.1/cpp_1.html#SEC1
'.
%
! ------------------- Instance methods for CPreprocessor
category: 'directives'
set compile_env: 0
method: CPreprocessor
addRestOfLineTo: aCollection

	[
		readStream atEnd or: [readStream peek == Character lf].
	] whileFalse: [
		aCollection add: self nextToken.
	].
%
category: 'other'
set compile_env: 0
method: CPreprocessor
addTokens: aList

	| lastToken |
	(tokens notEmpty and: [(lastToken := tokens last) isIdentifierToken and: [lastToken value last ~~ $( and: [aList first isOpenParenthesisToken]]]) ifTrue: [
		lastToken value add: $(.
		lastToken source add: aList first source.
		tokens addAll: (aList copyFrom: 2 to: aList size).
	] ifFalse: [
		tokens addAll: aList.
	].
%
category: 'other'
set compile_env: 0
method: CPreprocessor
applyConcatenationToken: aToken

	| left right list |
	left := tokens last.
	right := self nextToken.
	list := Array with: left with: aToken with: right.
	self applyToken: right.
	left 
		collectOriginalFrom: list;
		concatenate: tokens last.
	tokens size: tokens size - 1.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
applyDirectiveToken: aToken

	| directive |
	directive := aToken value.
	directive = 'define' 	ifTrue: [^self define: 	aToken].
	directive = 'undef' 	ifTrue: [^self undef: 	aToken].
	directive = 'ifdef' 		ifTrue: [^self ifdef: 	aToken defined: true].
	directive = 'ifndef' 	ifTrue: [^self ifdef: 	aToken defined: false].
	directive = 'if'			ifTrue: [^self if:			aToken].
	directive = 'include'	ifTrue: [^self include: aToken].
	self halt.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
applyFunctionMacroToken: aToken

	| name arguments definition oldDefinitions oldStream keys oldSize |
	name := aToken value.
	definition := definitions at: name.
	definitions removeKey: name.		"Avoid recursive definitions (e.g., #define a a)"
	arguments := self macroFunctionArgumentsFor: aToken.
	keys := Array new.
	oldDefinitions := KeyValueDictionary new.
	1 to: arguments size do: [:i | 
		| key value |
		key := keys add: (definition defineParameters at: i).
		value := CPreprocessorToken define: (arguments at: i).
		(definitions includesKey: key) ifTrue: [
			oldDefinitions at: key put: (definitions at: key).
		].
		definitions at: key put: value.
	].
	oldSize := tokens size.
	oldStream := readStream.
	readStream := ReadStream on: definition defineString.
	self readTillEnd.
	readStream := oldStream.
	keys do: [:each | 
		definitions removeKey: each.
	].
	oldDefinitions keysAndValuesDo: [:eachKey :eachValue | 
		definitions at: eachKey put: eachValue.
	].
	oldSize == tokens size ifTrue: [tokens add: CPreprocessorToken empty].
	(tokens at: oldSize + 1) replaceFirst: aToken.
	oldSize + 2 to: tokens size do: [:i | (tokens at: i) replaceOther: aToken].
	definitions at: name put: definition.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
applyIdentifierToken: aToken

	| name addOpenParenthesisTokenAtEnd oldTokens list |
	(tokens notEmpty and: [tokens last isDefinedIdentifierToken]) ifTrue: [tokens add: aToken. ^self].
	name := aToken value.
	(name last ~~ $( and: [readStream peek == $(]) ifTrue: [
		readStream next.
		aToken addOpenParenthesis.
	].
	addOpenParenthesisTokenAtEnd := false.
	"An identifier might include an open parenthesis to make definitions of macro functions easier to spot. We now need to remove it."
	((definitions includesKey: name) or: [name last == $( and: [
		addOpenParenthesisTokenAtEnd := true.
		name := name copyFrom: 1 to: name size - 1. 
		(definitions includesKey: name)]]) ifFalse: [tokens add: aToken. ^self].
	oldTokens := tokens.
	tokens := list := Array new.
	name last = $( ifTrue: [
		self applyFunctionMacroToken: aToken.
	] ifFalse: [
		self
			applySimpleMacroToken: aToken 
			name: name 
			addOpenParenthesisTokenAtEnd: addOpenParenthesisTokenAtEnd.
	].
	tokens := oldTokens.
	self addTokens: list.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
applySimpleMacroToken: aToken name: aString addOpenParenthesisTokenAtEnd: aBoolean

	| definition oldStream oldSize |
	"We need to do simple macro substitution"
	definition := definitions at: aString.
	definitions removeKey: aString.		"Avoid recursive definitions (e.g., #define a a)"
	oldSize := tokens size.
	oldStream := readStream.
	readStream := ReadStream on: definition defineString.
	self readTillEnd.
	readStream := oldStream.
	aBoolean ifTrue: [tokens add: (CPreprocessorToken punctuator: '(')].
	oldSize == tokens size ifTrue: [tokens add: CPreprocessorToken empty].
	(tokens at: oldSize + 1) replaceFirst: aToken.
	oldSize + 2 to: tokens size do: [:i | (tokens at: i) replaceOther: aToken].
	definitions at: aString put: definition.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
applyToken: aToken

	| lastToken |
	aToken isDirectiveToken ifTrue: [^self applyDirectiveToken: aToken].
	aToken isConcatenationToken ifTrue: [^self applyConcatenationToken: aToken].
	aToken isIdentifierToken ifTrue: [^self applyIdentifierToken: aToken].
	(aToken isStarToken and: [tokens notEmpty and: [(lastToken := tokens last) isIdentifierToken and: [lastToken value last == $(]]]) ifTrue: [
		lastToken removeOpenParenthesis.
		tokens add: (CPreprocessorToken punctuator: '(').
	].
	tokens add: aToken.
%
category: 'default definitions'
set compile_env: 0
method: CPreprocessor
defaultLinuxDefinitions
	"echo '' | cpp -dM"

	^'
#define __DBL_MIN_EXP__ (-1021)
#define __FLT_MIN__ 1.17549435e-38F
#define __CHAR_BIT__ 8
#define __WCHAR_MAX__ 2147483647
#define __GCC_HAVE_SYNC_COMPARE_AND_SWAP_1 1
#define __GCC_HAVE_SYNC_COMPARE_AND_SWAP_2 1
#define __GCC_HAVE_SYNC_COMPARE_AND_SWAP_4 1
#define __DBL_DENORM_MIN__ 4.9406564584124654e-324
#define __GCC_HAVE_SYNC_COMPARE_AND_SWAP_8 1
#define __FLT_EVAL_METHOD__ 0
#define __unix__ 1
#define __x86_64 1
#define __DBL_MIN_10_EXP__ (-307)
#define __FINITE_MATH_ONLY__ 0
#define __GNUC_PATCHLEVEL__ 3
#define __DEC64_MAX_EXP__ 385
#define __SHRT_MAX__ 32767
#define __LDBL_MAX__ 1.18973149535723176502e+4932L
#define __UINTMAX_TYPE__ long unsigned int
#define __linux 1
#define __DEC32_EPSILON__ 1E-6DF
#define __unix 1
#define __LDBL_MAX_EXP__ 16384
#define __linux__ 1
#define __SCHAR_MAX__ 127
#define __DBL_DIG__ 15
#define _FORTIFY_SOURCE 2
#define __SIZEOF_INT__ 4
#define __SIZEOF_POINTER__ 8
#define __USER_LABEL_PREFIX__ 
#define __STDC_HOSTED__ 1
#define __LDBL_HAS_INFINITY__ 1
#define __FLT_EPSILON__ 1.19209290e-7F
#define __LDBL_MIN__ 3.36210314311209350626e-4932L
#define __DEC32_MAX__ 9.999999E96DF
#define __SIZEOF_LONG__ 8
#define __DECIMAL_DIG__ 21
#define __gnu_linux__ 1
#define __LDBL_HAS_QUIET_NAN__ 1
#define __GNUC__ 4
#define __MMX__ 1
#define __FLT_HAS_DENORM__ 1
#define __SIZEOF_LONG_DOUBLE__ 16
#define __BIGGEST_ALIGNMENT__ 16
#define __DBL_MAX__ 1.7976931348623157e+308
#define __DBL_HAS_INFINITY__ 1
#define __DEC32_MIN_EXP__ (-94)
#define __LDBL_HAS_DENORM__ 1
#define __DEC128_MAX__ 9.999999999999999999999999999999999E6144DL
#define __DEC32_MIN__ 1E-95DF
#define __DBL_MAX_EXP__ 1024
#define __DEC128_EPSILON__ 1E-33DL
#define __SSE2_MATH__ 1
#define __amd64 1
#define __LONG_LONG_MAX__ 9223372036854775807LL
#define __SIZEOF_SIZE_T__ 8
#define __SIZEOF_WINT_T__ 4
#define __GCC_HAVE_DWARF2_CFI_ASM 1
#define __GXX_ABI_VERSION 1002
#define __FLT_MIN_EXP__ (-125)
#define __DBL_MIN__ 2.2250738585072014e-308
#define __LP64__ 1
#define __DECIMAL_BID_FORMAT__ 1
#define __DEC128_MIN__ 1E-6143DL
#define __REGISTER_PREFIX__ 
#define __DBL_HAS_DENORM__ 1
#define __NO_INLINE__ 1
#define __FLT_MANT_DIG__ 24
#define __VERSION__ "4.4.3"
#define __DEC64_EPSILON__ 1E-15DD
#define __DEC128_MIN_EXP__ (-6142)
#define unix 1
#define __SIZE_TYPE__ long unsigned int
#define __ELF__ 1
#define __FLT_RADIX__ 2
#define __LDBL_EPSILON__ 1.08420217248550443401e-19L
#define __SSE_MATH__ 1
#define __k8 1
#define __SIZEOF_PTRDIFF_T__ 8
#define __x86_64__ 1
#define __DEC32_SUBNORMAL_MIN__ 0.000001E-95DF
#define __FLT_HAS_QUIET_NAN__ 1
#define __FLT_MAX_10_EXP__ 38
#define __LONG_MAX__ 9223372036854775807L
#define __DEC128_SUBNORMAL_MIN__ 0.000000000000000000000000000000001E-6143DL
#define __FLT_HAS_INFINITY__ 1
#define __DEC64_MAX__ 9.999999999999999E384DD
#define __CHAR16_TYPE__ short unsigned int
#define __DEC64_MANT_DIG__ 16
#define __DEC32_MAX_EXP__ 97
#define linux 1
#define __SSE2__ 1
#define __LDBL_MANT_DIG__ 64
#define __DBL_HAS_QUIET_NAN__ 1
#define __k8__ 1
#define __WCHAR_TYPE__ int
#define __SIZEOF_FLOAT__ 4
#define __DEC64_MIN_EXP__ (-382)
#define __FLT_DIG__ 6
#define __INT_MAX__ 2147483647
#define __amd64__ 1
#define __FLT_MAX_EXP__ 128
#define __DBL_MANT_DIG__ 53
#define __DEC64_MIN__ 1E-383DD
#define __WINT_TYPE__ unsigned int
#define __SIZEOF_SHORT__ 2
#define __SSE__ 1
#define __LDBL_MIN_EXP__ (-16381)
#define __SSP__ 1
#define __LDBL_MAX_10_EXP__ 4932
#define __DBL_EPSILON__ 2.2204460492503131e-16
#define _LP64 1
#define __SIZEOF_WCHAR_T__ 4
#define __DEC_EVAL_METHOD__ 2
#define __INTMAX_MAX__ 9223372036854775807L
#define __FLT_DENORM_MIN__ 1.40129846e-45F
#define __CHAR32_TYPE__ unsigned int
#define __FLT_MAX__ 3.40282347e+38F
#define __SIZEOF_DOUBLE__ 8
#define __FLT_MIN_10_EXP__ (-37)
#define __INTMAX_TYPE__ long int
#define __DEC128_MAX_EXP__ 6145
#define __GNUC_MINOR__ 4
#define __DEC32_MANT_DIG__ 7
#define __DBL_MAX_10_EXP__ 308
#define __LDBL_DENORM_MIN__ 3.64519953188247460253e-4951L
#define __STDC__ 1
#define __PTRDIFF_TYPE__ long int
#define __DEC64_SUBNORMAL_MIN__ 0.000000000000001E-383DD
#define __DEC128_MANT_DIG__ 34
#define __LDBL_MIN_10_EXP__ (-4931)
#define __SIZEOF_LONG_LONG__ 8
#define __LDBL_DIG__ 18
#define __GNUC_GNU_INLINE__ 1
'.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
defaultSearchPaths

	| list |
	list := OrderedCollection with: ''.
	(path notNil and: [path first = $$]) ifTrue: [list add: self directory].
	^list
		add: '/usr/include/';
		add: '/usr/local/include/';
		add: '/usr/include/linux/';
		add: '/usr/include/c++/4.4/tr1/';
		add: '/usr/include/c++/4.4/';
		yourself.
%
category: 'default definitions'
set compile_env: 0
method: CPreprocessor
defaultSparcSolarisDefinitions
	"echo '' | cpp -dM"

	^'
#define __GCC_NEW_VARARGS__ 1 
#define __sparc 1 
#define __svr4__ 1 
#define __sun 1 
#define sparc 1 
#define __sun__ 1 
#define __unix 1 
#define __unix__ 1 
#define __SVR4 1 
#define sun 1 
#define __sparc__ 1 
#define unix 1 
'.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
define: defineToken

	| nameToken name parameters list string |
	readStream peek == Character lf ifTrue: [self syntaxError].
	(nameToken := self nextToken) isIdentifierToken ifFalse: [self syntaxError].
	name := nameToken value.
	definitions at: name put: defineToken.
	list := Array with: defineToken with: nameToken.
	name last == $( ifTrue: [
		| token |
		parameters := Array new.
		token := list add: self nextToken.
		[
			token isCloseParenthesisToken.
		] whileFalse: [
			parameters add: token value.
			(token := list add: self nextToken) isCommaToken ifTrue: [
				token := list add: self nextToken.
			].
		].
	].
	string := String new.
	[
		readStream atEnd or: [readStream peek == Character lf].
	] whileFalse: [
		| token |
		token := self nextToken.
		string add: token source.
	].
	defineToken collectOriginalFrom: list.
	defineToken source add: string.
	defineToken
		defineParameters: parameters
		string: string.
%
category: 'accessors'
set compile_env: 0
method: CPreprocessor
definitions

	^definitions.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
directory

	path isNil ifTrue: [^''].
	path size to: 1 by: -1 do: [:i | 
		(path at: i) = $/ ifTrue: [
			^path copyFrom: 1 to: i.
		].
	].
	^''.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
doIf: aBoolean

	| inTrueBlock anyWereTrue token |
	inTrueBlock := aBoolean.
	anyWereTrue := false.
	[true] whileTrue: [
		inTrueBlock ifTrue: [
			anyWereTrue := true.
			[
				(token := self nextToken) isConditionalEndDirective.
			] whileFalse: [
				self applyToken: token.
			].
			(readStream atEnd or: [readStream peek == Character lf or: [token value = 'elif']]) ifFalse: [self syntaxError].
			token value = 'endif' ifTrue: [
				self ignoreRestOfLine.
				^self.
			].
			token value = 'else' ifTrue: [
				self ignoreRestOfLine.
				inTrueBlock := false.
			] ifFalse: [
				self ignoreRestOfLine.
				(anyWereTrue and: [token value = 'elif']) ifTrue: [
					inTrueBlock := false.
				] ifFalse: [
					self halt.
				].
			].
		] ifFalse: [		"not inTrueBlock"
			| ifDepth |
			ifDepth := 0.
			token := self skipWhile: [:each | 
				each isConditionalBeginDirective ifTrue: [ifDepth := ifDepth + 1].
				0 < ifDepth ifTrue: [
					each isEndifDirective ifTrue: [ifDepth := ifDepth - 1].
					true.
				] ifFalse: [
					each isConditionalEndDirective not.
				].
			].
			(readStream atEnd or: [readStream peek == Character lf or: [token value = 'if' or: [token value = 'elif']]]) ifFalse: [self syntaxError].
			token value = 'endif' ifTrue: [^self].
			token value = 'else' ifTrue: [
				inTrueBlock := anyWereTrue not.
			] ifFalse: [
				token value = 'if' ifTrue: [
					self skipWhile: [:each | each isEndifDirective not].
					readStream peek == Character lf ifFalse: [self syntaxError].
				] ifFalse: [
					token value = 'elif' ifTrue: [
						inTrueBlock := self evaluateIfOrElif: token.
					] ifFalse: [
						self halt.
					].
				].
			].
		].
	].
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
evaluate: aStream
	"Returns an Integer from a stream of tokens"

	| expr value |
	expr := self parse: aStream.
	value := self valueOf: expr.
	^value.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
evaluateHeader: aCHeader

	header := aCHeader.
	definitions := aCHeader types.
	^self evaluate: header.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
evaluateIfOrElif: aToken

	| oldTokens stream expression value |
	oldTokens := tokens.
	tokens := Array with: aToken.
	[
		readStream atEnd or: [readStream peek == Character lf].
	] whileFalse: [
		self applyToken: self nextToken.
	].
	stream := ReadStream on: tokens.
	tokens := oldTokens.
	stream next.
	expression := self parse: stream.
	value := self valueOf: expression.
	^value ~= 0.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
if: aToken

	self doIf: (self evaluateIfOrElif: aToken).
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
ifdef: aToken defined: aBoolean

	| nameToken name |
	readStream peek == Character lf ifTrue: [self syntaxError].
	(nameToken := self nextToken) isIdentifierToken ifFalse: [self syntaxError].
	name := nameToken value.
	readStream peek == Character lf ifFalse: [self syntaxError].
	self doIf: (definitions includesKey: name) == aBoolean.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
ignoreRestOfLine

	self addRestOfLineTo: OrderedCollection new.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
include: aToken

	| token string |
	token := self nextToken.
	aToken source add: token source.
	token isStringToken ifTrue: [
		string := self directory , token value.
		(GsFile existsOnServer: string) ifTrue: [^self includePath: string].
		self error: 'Include file not found!'.
	].
	token isOpenAngleBracketToken ifFalse: [self syntaxError].
	string := String new.
	[
		(token := self nextToken) isCloseAngleBracketToken.
	] whileFalse: [
		string add: token source.	"we want the source rather than the valueString since 'gnu/stubs-64.h' ends up with an integer token of '64.h'!"
		aToken source add: token source.
	].
	searchPaths do: [:each | 
		| fullPath |
		fullPath := each , string.
		(GsFile existsOnServer: fullPath) ifTrue: [^self includePath: fullPath].
	].
	(GsFile existsOnServer: self directory , string) ifTrue: [^self includePath: self directory , string].
	self error: 'Include file not found!'.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
includePath: aString

	| oldPath oldLine oldStream |
	oldPath := path.
	oldLine := line.
	oldStream := readStream.
	path := aString.
	line := 1.
	(readStream := GsFile openReadOnServer: aString) isNil ifTrue: [self error: GsFile serverErrorString].
	[
		self readTillEnd.
	] ensure: [
		readStream close.
		readStream := oldStream.
		line := oldLine.
		path := oldPath.
	].
%
category: 'other'
set compile_env: 0
method: CPreprocessor
initialize

	tokens := OrderedCollection new.
	searchPaths := self defaultSearchPaths.
	self initializeDefinitions.
%
category: 'default definitions'
set compile_env: 0
method: CPreprocessor
initializeDefinitions

	| gsBuildArchitecture string |
	gsBuildArchitecture := System gemVersionReport at: 'gsBuildArchitecture'.
	definitions := self class defaultDefinitions
		at: gsBuildArchitecture
		ifAbsentPut: [
			definitions := KeyValueDictionary new.
			string := 
				gsBuildArchitecture = 'x86-64 (Linux)' ifTrue: [self defaultLinuxDefinitions] ifFalse: [
				"gsBuildArchitecture = 'SPARC (Solaris)' ifTrue: [self defaultSparcSolarisDefinitions] ifFalse: ["
				self error: gsBuildArchitecture , ' build type is not (yet?) supported'].
			self parseString: string.
			definitions.
		].
	definitions := definitions copy.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
integerFromBoolean: aBoolean

	^aBoolean 
		ifTrue: [1] 
		ifFalse: [0].
%
category: 'other'
set compile_env: 0
method: CPreprocessor
macroFunctionArgumentsFor: aToken

	| arguments index parenthesisCount list token |
	arguments := Array new.
	parenthesisCount := 1.
	list := Array with: aToken.
	index := 2.
	[
		token := self nextToken.
		(parenthesisCount == 1 and: [token isCommaToken]) ifTrue: [arguments add: (list copyFrom: index to: list size). index := list size + 2] ifFalse: [
		token isOpenParenthesisToken ifTrue: [parenthesisCount := parenthesisCount + 1] ifFalse: [
		token isCloseParenthesisToken ifTrue: [parenthesisCount := parenthesisCount - 1]]].
		list add: token.
		0 < parenthesisCount.
	] whileTrue: [].
	arguments add: (list copyFrom: index to: list size - 1).
	aToken collectOriginalFrom: list.
	arguments := arguments collect: [:eachArray | 
		| string oldStream oldTokens |
		string := String new.
		eachArray do: [:each | string add: each source].
		oldStream := readStream.
		readStream := ReadStream on: string.
		oldTokens := tokens.
		tokens := Array new.
		self readTillEnd.
		string := String new.
		tokens do: [:each | string add: each valueString; add: Character space].
		readStream := oldStream.
		tokens := oldTokens.
		string.
	].
	^arguments.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
mergeLastTwoElementsIn: stack

	| x y token |
	y := stack removeLast.
	x := stack last.
	token := (y at: 2) -> (Array with: (x at: 3) with: (y at: 3)).
	x at: 3 put: token.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
nextToken
		"This can return nil if there are no more tokens.
		The stream might not be atEnd because of trailing comments."

	| token |
	token := CPreprocessorToken
		nextFrom: readStream 
		filename: path 
		line: line.
	token notNil ifTrue: [line := token line].
	^token.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
numberFrom: aString
	"Returns an number literal token"

	| string value isLong isUnsigned |
	isLong := false.
	isUnsigned := false.
	(2 < aString size and: [(aString at: 1) = $0 and: [(aString at: 2) = $x]]) ifTrue: [
		string := '16r' , (aString copyFrom: 3 to: aString size).
	] ifFalse: [
		aString first == $0 ifTrue: [
			string := '8r' , aString.
		] ifFalse: [
			string := aString.
		].
	].
	('lL' includes: string last) ifTrue: [
		isLong := true.
		string := string copyFrom: 1 to: string size - 1.
	].
	('uU' includes: string last) ifTrue: [
		isUnsigned := true.
		string := string copyFrom: 1 to: string size - 1.
		('lL' includes: string last) ifTrue: [
			isLong := true.
			string := string copyFrom: 1 to: string size - 1.
		].
	].
	value := Integer fromString: string.
	(value class ~~ SmallInteger and: [16rFFFFFFFFFFFFFFFF < value]) ifTrue: [self halt].
	(isLong or: [16rFFFFFFFF < value]) ifTrue: [
		(isUnsigned or: [16r7FFFFFFFFFFFFFFF < value]) 
			ifTrue: [^CPreprocessorToken uint64: value]
			ifFalse: [^CPreprocessorToken int64: value].
	].
	16rFFFF < value ifTrue: [
		(isUnsigned or: [16r7FFFFFFF < value]) 
			ifTrue: [^CPreprocessorToken uint32: value]
			ifFalse: [^CPreprocessorToken int32: value].
	].
	^(isUnsigned or: [16r7FFF < value]) 
		ifTrue: [CPreprocessorToken uint16: value]
		ifFalse: [CPreprocessorToken int16: value].
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
numberLiteralKeys

	^#(
		#'int8' #'uint8' #'int16' #'uint16' #'int32' #'uint32' #'int64' #'uint64'
	).
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
operandFrom: aStream
	"Returns a token"

	| token |
	token := aStream next.
	token key = #'integer' ifTrue: [^token].
	token key = #'number' ifTrue: [^self numberFrom: token value].
	token key = #'defined' ifTrue: [^#'integer' -> (self integerFromBoolean: (definitions includesKey: token value))].
	token key = #'punctuator' ifTrue: [
		token value = '!' ifTrue: [
			token := self operandFrom: aStream.
			token value: (token value == 1 ifTrue: [0] ifFalse: [1]).
			^token.
		].
		token value = '(' ifTrue: [
			| value |
			value := self evaluate: aStream.
			token := aStream next.
			(token key = #'punctuator' and: [token value = ')']) ifTrue: [^#'integer' -> value].
self halt.
		].
self halt.
	].
	token key = #'identifier' ifTrue: [
		| definition |
		definition := definitions
			at: token value
			ifAbsent: [^#'integer' -> 0].
		definition key isNil ifTrue: [
			^#'integer' -> (self evaluate: (ReadStream on: definition value)).
		] ifFalse: [
			self halt.
		].
	].
self halt.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
parse: aStream
	"Returns a token from a stream of tokens"

	| list stack |
	list := self parseA: aStream.		"Each element is a three-element array: { precedence. operator. operand }"
	stack := OrderedCollection with: list removeFirst.
	1 to: list size do: [:i | 
		| each |
		each := list at: i.
		[
			stack last first >= each first.
		] whileTrue: [
			self mergeLastTwoElementsIn: stack.
		].
		stack add: each.
	].
	[
		1 < stack size.
	] whileTrue: [
		self mergeLastTwoElementsIn: stack.
	].
	^stack first last.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
parseA: aStream
	"Returns a list of the following: { precedence. operator. operand }"

	| list |
	list := OrderedCollection with: (Array
		with: -999
		with: ''
		with: (self parseForOperandFrom: aStream)).
	[true] whileTrue: [
		| operand operator precedence |
		(operator := aStream peek) isNil ifTrue: [^list].
		operator key == #'punctuator' ifFalse: [^list].
		(precedence := self precedenceOf: operator value) isNil ifTrue: [^list].
		aStream next.
		list add: (Array 
			with: precedence 
			with: operator value asSymbol
			with: (self parseForOperandFrom: aStream)).
	].
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
parseForDefinedFrom: aStream expectCloseParenthesis: aBoolean

	| identifierToken closeParenthesisToken key found |
	(identifierToken := aStream next) isIdentifierToken ifFalse: [self syntaxError].
	aBoolean ifTrue: [(closeParenthesisToken := aStream next) isCloseParenthesisToken ifFalse: [self syntaxError]].
	key := identifierToken value.
	found := definitions includesKey: identifierToken value.
	^found 
		ifTrue: [#'int16' -> 1]
		ifFalse: [#'int16' -> 0].
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
parseForFunction: aString from: aStream

	aString = 'sizeof(' ifTrue: [^self parseForSizeofFrom: aStream].
	self halt.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
parseForOperandFrom: aStream
	"Returns an expression token (which is different from a preprocessor token?)"

	| token |
	token := aStream next.
	(self numberLiteralKeys includes: token key) ifTrue: [^token].
	token key == #'integer' 		ifTrue: [^token].
	token key == #'identifier' 		ifTrue: [
		aStream atEnd ifTrue: [^token].
		token isDefinedIdentifierToken ifTrue: [^self parseForDefinedFrom: aStream expectCloseParenthesis: token value last == $(].
		token value last == $( ifFalse: [^token].
		^self parseForFunction: token value from: aStream.
	].
	token key == #'number' 		ifTrue: [^self numberFrom: token value].
	token key == #'defined' 		ifTrue: [^token].
	token key == #'empty' 			ifTrue: [^token].
	token key == #'character'		ifTrue: [^token].
	token key == #'punctuator' 	ifTrue: [
		token value = '!' 				ifTrue: [
			token := self parseForOperandFrom: aStream.
			^#'not' -> token.
		].
		token value = '-' 				ifTrue: [
			token := self parseForOperandFrom: aStream.
			^#'negate' -> token.
		].
		token value = '(' 				ifTrue: [
			| typeCast endToken |
			(typeCast := self readTypeSpecifierFrom: aStream) isNil ifTrue: [
				token := self parse: aStream.
			].
			(endToken := aStream next) isCloseParenthesisToken ifFalse: [self error: 'Expected close parenthesis'].
			typeCast notNil ifTrue: [
				token := self parseForOperandFrom: aStream.
				token key == typeCast ifTrue: [^token].
				"self halt."
			].
			^token.
		].
	].
	self error: 'unhandled expression token!'.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
parseForSizeofFrom: aStream

	| list token myHeader oldTokens declarations declaration |
	list := OrderedCollection new.
	[
		(token := aStream next) isCloseParenthesisToken.
	] whileFalse: [
		list add: token.
	].
	list first isStructIdentifierToken ifTrue: [^self parseForSizeofStructFrom: list].
	oldTokens := tokens.
	tokens := list.
	myHeader := (System myUserProfile symbolList objectNamed: #'CHeader') preprocessor: self.
	tokens := oldTokens.
	(declarations := myHeader declarations) size == 1 ifFalse: [self error: 'Expected exactly one declaration'].
	declaration := declarations first.
	declaration name notNil ifTrue: [
		declaration := definitions at: declaration name.
	].
	^#'int32' -> declaration byteSize.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
parseForSizeofStructFrom: aList

	| token fields size |
	aList size = 2 ifFalse: [self error: 'Unexpected structure!'].
	(token := aList at: 2) key == #'identifier' ifFalse: [self error: 'Expected an identifier'].
	fields := header fieldsForStruct: token value.
	size := fields
		inject: 0
		into: [:sum :each | sum + each byteSize].
	^CPreprocessorToken int32: size.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
parseString: aString

	readStream := ReadStream on: aString.
	line := 1.
	self readTillEnd.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
precedenceOf: aString
	"Answer a negative number so that * (-3) is GREATER than + (-4) "

	(#('*' '/' '%') includes: aString) 			ifTrue: [^-3].
	(#('+' '-') includes: aString) 				ifTrue: [^-4].
	(#('<<' '>>') includes: aString)				ifTrue: [^-5].
	(#('<' '<=' '>' '>=') includes: aString) 	ifTrue: [^-6].
	(#('==' '!=') includes: aString) 				ifTrue: [^-7].
	'&' = aString										ifTrue: [^-8].
	'^' = aString										ifTrue: [^-9].
	'|' = aString										ifTrue: [^-10].
	'&&' = aString										ifTrue: [^-11].
	'||' = aString										ifTrue: [^-12].
	^nil.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
readTillEnd

	| token |
	[
		(token := self nextToken) notNil.
	] whileTrue: [
		self applyToken: token.
	].
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
readTypeSpecifierFrom: aStream
 
	| token value signed size |
	(token := aStream peek) key == #'identifier' ifFalse: [^nil].
	value := token value.
	(definitions includesKey: value) ifTrue: [aStream next. ^(definitions at: value) type].
	value = 'struct'	ifTrue: [^nil].
	value = 'void' 		ifTrue: [aStream next. ^aStream peek isStarToken ifTrue: [aStream next. #'ptr'] ifFalse: [#'void']].
	value = 'size_t' 	ifTrue: [aStream next. ^#'uint32'	].
	value = 'float' 		ifTrue: [aStream next. ^#'float'		].
	value = 'double' 	ifTrue: [aStream next. ^#'double'	].
	(#('signed' 'unsigned') includes: value) ifTrue: [
		signed := value = 'signed'.
		(token := aStream next; peek) key == #'identifier' ifFalse: [self error: 'Expected an identifier'].
		value := token value.
	].
	value = 'char' ifTrue: [
		aStream next. 
		^signed == true
		ifTrue: [#'int8']
		ifFalse: [#'uint8'].
	].
	(#('long' 'short') includes: value) ifTrue: [
		size := value.
		(token := aStream next; peek) key == #'identifier' ifFalse: [self error: 'Expected an identifier'].
		value := token value.
	].
	value = 'double' ifTrue: [
		aStream next. 
		size = 'long' ifFalse: [self error: 'Unexpected type']. 
		^#'longDouble'.
	].
	value = 'int' ifTrue: [
		aStream next. 
		^self typeForIntSigned: signed size: size.
	].
	size notNil ifTrue: [
		^self typeForIntSigned: signed size: size.
	].
	signed notNil ifTrue: [self error: 'Unexpected type'].
	(signed isNil and: [size isNil]) ifTrue: [^nil].
	^self 
		typeForIntSigned: signed 
		size: size.
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
skipWhile: aBlock

	| token |
	[
		token := self nextToken.
		aBlock value: token.
	] whileTrue: [].
	^token.
%
category: 'other'
set compile_env: 0
method: CPreprocessor
syntaxError

	self error: 'Syntax error!'.
%
category: 'accessors'
set compile_env: 0
method: CPreprocessor
tokens

	^tokens.
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
typeForIntSigned: aBooleanOrNil size: aStringOrNil
"
// From Linux, we have the following: 
#define __SIZEOF_SHORT__ 2
#define __SIZEOF_INT__ 4
#define __SIZEOF_LONG__ 8
#define __SIZEOF_LONG_LONG__ 8
"
	aStringOrNil = 'long' ifTrue: [
		^aBooleanOrNil == false 
			ifTrue: [#'uint64']
			ifFalse: [#'int64']. 
	].
	aStringOrNil = 'short' ifTrue: [
		^aBooleanOrNil == false 
			ifTrue: [#'uint16']
			ifFalse: [#'int16']. 
	].
	^aBooleanOrNil == false 
		ifTrue: [#'uint32']
		ifFalse: [#'int32'].
%
category: 'directives'
set compile_env: 0
method: CPreprocessor
undef: defineToken

	| nameToken name |
	readStream peek == Character lf ifTrue: [self syntaxError].
	(nameToken := self nextToken) isIdentifierToken ifFalse: [self syntaxError].
	readStream peek == Character lf ifFalse: [self syntaxError].
	name := nameToken value.
name = 'XXX' ifTrue: [self halt].
	definitions 
		removeKey: name
		ifAbsent: [].
%
category: 'expression evaluation'
set compile_env: 0
method: CPreprocessor
valueOf: aToken
	"Returns an Integer from a token."

	| x y |
	(self numberLiteralKeys includes: aToken key)		ifTrue: [^aToken value].
	aToken key == #'character'	ifTrue: [^aToken value codePoint].
	aToken key == #'empty'		ifTrue: [^0].
	aToken key == #'identifier' 	ifTrue: [
		| assoc |
		assoc := definitions 
			at: aToken value 
			ifAbsent: [nil].
		assoc isNil ifTrue: [
			header isNil ifTrue: [^0].
			^header enums 
				at: aToken value 
				ifAbsent: [0].
		].
		^self evaluate: (ReadStream on: assoc value)].
	aToken key == #'defined' 		ifTrue: [^self integerFromBoolean: (definitions includesKey: aToken value)].
	aToken key == #'not'			ifTrue: [^self integerFromBoolean: (self valueOf: aToken value) = 0].
	aToken key == #'negate'		ifTrue: [^(self valueOf: aToken value) negated].
	aToken key == #'&&' 			ifTrue: [
		(self valueOf: (aToken value at: 1)) = 0 ifTrue: [^0].
		^self valueOf: (aToken value at: 2). 
	].
	aToken key == #'||' 			ifTrue: [
		(self valueOf: (aToken value at: 1)) = 1 ifTrue: [^1].
		^self valueOf: (aToken value at: 2). 
	].
	x := self valueOf: (aToken value at: 1).
	y := self valueOf: (aToken value at: 2).
	aToken key == #'=='			ifTrue: [^self integerFromBoolean: x = y].
	aToken key == #'>='			ifTrue: [^self integerFromBoolean: x >= y].
	aToken key == #'>'				ifTrue: [^self integerFromBoolean: x > y].
	aToken key == #'<'				ifTrue: [^self integerFromBoolean: x < y].
	aToken key == #'<='			ifTrue: [^self integerFromBoolean: x <= y].
	aToken key == #'*'				ifTrue: [^x * y].
	aToken key == #'+'				ifTrue: [^x + y].
	aToken key == #'/'				ifTrue: [^x // y].
	aToken key == #'-'				ifTrue: [^x - y].
	aToken key == #'<<'			ifTrue: [^x bitShift: y].
	aToken key == #'>>'			ifTrue: [^x bitShift: y negated].

	self halt.
%
run
"Initialize class variable and allow it to be modified by any user"
CPreprocessor defaultDefinitions.
true.
%
