! ========================================================================
! Copyright (C) by VMware, Inc. 1991-2011.  All Rights Reserved
! ========================================================================

expectvalue %String
run
(Object
	subclass: 'CPreprocessorToken'
	instVarNames: #( file line source
                    type value)
	classVars: #()
	classInstVars: #()
	poolDictionaries: { }
	inDictionary: Globals
	options: #( #logCreation) )
definition
%

! set class category
run
CPreprocessorToken category: 'FFI'.
true
%

! Remove existing behavior from CPreprocessorToken
doit
CPreprocessorToken removeAllMethods.
CPreprocessorToken class removeAllMethods.
true
%
! ------------------- Class methods for CPreprocessorToken
category: 'Documentation'
set compile_env: 0
classmethod: CPreprocessorToken
comment
^'This class is internal to the FFI implementation and 
represents a preprocessor token from a C header file.'
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
args: anArray
	| res |
	(res := self basicNew) initializeArgs: anArray .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
concatenation
	| res |
	(res := self basicNew) initializeConcatenation .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
define: anArray
	| res |
	(res := self basicNew) initializeDefine: anArray .
    ^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
directive: aString
	| res |
	(res := self basicNew) initializeDirective: aString .
    ^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
empty
	| res |
	(res := self basicNew) initializeEmpty .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
identifier: aString
	| res |
	(res := self basicNew) initializeIdentifier: aString .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
int16: anInteger
	| res |
	(res := self basicNew ) initializeInt16: anInteger .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
int32: anInteger
	| res |
	(res := self basicNew) initializeInt32: anInteger .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
int64: anInteger
	| res |
	(res := self basicNew) initializeInt64: anInteger .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
nextFrom: aStream filename: aString line: anInteger
		"Note that the initialize method might return nil if no more tokens!"
	^ self basicNew
		initializeStream: aStream 
		filename: aString 
		line: anInteger .
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
number: aNumber
	| res |
	(res := self basicNew) initializeNumber: aNumber .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
punctuator: aString
	| res |
	(res := self basicNew) initializePunctuator: aString .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
string: aString
	| res |
	(res := self basicNew) initializeString: aString .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
uint16: anInteger
	| res |
	(res := self basicNew) initializeUInt16: anInteger .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
uint32: anInteger
	| res |
	(res := self basicNew) initializeUInt32: anInteger .
	^ res
%
category: 'other'
set compile_env: 0
classmethod: CPreprocessorToken
uint64: anInteger
	| res |
	(res := self basicNew) initializeUInt64: anInteger.
	^ res
%
! ------------------- Instance methods for CPreprocessorToken
category: 'other'
set compile_env: 0
method: CPreprocessorToken
addOpenParenthesis

	value add: $(.
	source add: $(.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
asAssociation
	"To facilitate automated tests"
	^type -> value.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
collectOriginalFrom: aList

	| string |
	string := String new.
	aList do: [:each | string add: each source].
	source := string.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
concatenate: aToken

	value add: (
		aToken isIdentifierToken ifTrue: [aToken value] ifFalse: [
		aToken isIntegerToken ifTrue: [aToken value printString] ifFalse: [
		self halt]]).
%
category: 'Accessing'
set compile_env: 0
method: CPreprocessorToken
defineParameters

	^value key.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
defineParameters: anArray string: aString

	type := #'define'.
	value := anArray -> aString.
%
category: 'Accessing'
set compile_env: 0
method: CPreprocessorToken
defineString

	^value value.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
evaluateExpression

self halt.
	type == #'int16' ifTrue: [^value].
	type == #'identifier' ifTrue: [^0].	"#if FOO // undefined identifier is equivalent to false"
	self isEmptyToken ifTrue: [self error: 'Empty expression'].
	self halt.
%
category: 'Accessing'
set compile_env: 0
method: CPreprocessorToken
file

	^file.
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
hexCharacterFrom: aStream
	"Enter with pointing to 'x' of '\0xhh'"

	| string |
	string := String withAll: '16r0'.
	[true] whileTrue: [
		| char |
		char := self peekFrom: aStream.
		(string size < 6 and: [char isAlphaNumeric]) ifFalse: [
			value add: (Character codePoint: (Integer fromString: string)).
			^self.
		].
		string add: char.
		source add: aStream next.
	].
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeArgs: anArray

	type := #'args'.
	value := anArray.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeConcatenation

	type := #'concatenation'.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeDefine: anArray

	type := #'define'.
	value := nil -> anArray.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeDirective: aString

	type := #'directive'.
	value := aString.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeEmpty

	type := #'empty'.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeIdentifier: aString

	type := #'identifier'.
	value := aString.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeInt16: anInteger

	type := #'int16'.
	value := anInteger.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeInt32: anInteger

	type := #'int32'.
	value := anInteger.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeInt64: anInteger

	type := #'int64'.
	value := anInteger.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeNumber: aNumber

	type := #'number'.
	value := aNumber.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializePunctuator: aString

	type := #'punctuator'.
	value := aString.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
initializeStream: aStream filename: aString line: anInteger
		"The return value from the class-side method will be what this method returns, not always a new object"

	| isBeginningOfLine char1 char2 |
	isBeginningOfLine := self isBeginningOfLineFor: aStream.
	file := aString.
	line := anInteger.
	source := String new.
	[
		[
			self skipCommentsIn: aStream.
		] whileTrue: [
			line := line + 1.
			source add: aStream next.
		].
		aStream peek == $\.
	] whileTrue: [
		source add: aStream next.
	].
	(char1 := self peekFrom: aStream) isNil ifTrue: [^nil].
	char2 := aStream peek2.
	(char1 == $L and: [char2 == $']) ifTrue: [self readWideCharacterFrom: aStream] ifFalse: [
	(char1 == $L and: [char2 == $"]) ifTrue: [self readWideStringFrom: aStream] ifFalse: [
	char1 isLetter	ifTrue: [	self readIdentifierFrom: aStream] ifFalse: [
	char1 == $_	ifTrue: [	self readIdentifierFrom: aStream] ifFalse: [
	char1 == $" 	ifTrue: [	self readStringFrom: aStream] ifFalse: [
	char1 == $' 	ifTrue: [	self readCharacterFrom: aStream] ifFalse: [
	char1 == $# 	ifTrue: [	self readHashFrom: aStream isBeginningOfLine: isBeginningOfLine] ifFalse: [
	char1 isDigit 	ifTrue: [	self readNumberFrom: aStream] ifFalse: [
									self readPunctuatorFrom: aStream]]]]]]]].
	"Include trailing comments starting on this line with the source for this token"
	self skipCommentsIn: aStream.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeString: aString

	type := #'string'.
	value := aString.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeUInt16: anInteger

	type := #'uint16'.
	value := anInteger.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeUInt32: anInteger

	type := #'uint32'.
	value := anInteger.
%
category: 'Initializing'
set compile_env: 0
method: CPreprocessorToken
initializeUInt64: anInteger

	type := #'uint64'.
	value := anInteger.
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isAttributeIdentifierToken

	^type == #'identifier' and: [value = '__attribute__('].
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
isBeginningOfLineFor: aStream

	aStream atBeginning ifTrue: [^true].
	^aStream peek == Character lf.
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isCloseAngleBracketToken

	^type == #'punctuator' and: [value = '>'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isCloseCurlyBracketToken

	^type == #'punctuator' and: [value = '}'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isCloseParenthesisToken

	^type == #'punctuator' and: [value = ')'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isCloseSquareBracketToken

	^type == #'punctuator' and: [value = ']'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isColonToken

	^type == #'punctuator' and: [value = ':'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isCommaToken

	^type == #'punctuator' and: [value = ','].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isConcatenationToken

	^type == #'concatenation'.
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isConditionalBeginDirective

	| list |
	list := #('if' 'ifdef' 'ifndef').
	^type == #'directive' and: [list includes: value].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isConditionalDirective

	| list |
	list := #('if' 'else' 'endif' 'ifdef' 'ifndef' 'elif').
	^type == #'directive' and: [list includes: value].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isConditionalEndDirective

	| list |
	list := #('else' 'endif' 'elif').
	^type == #'directive' and: [list includes: value].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isDefinedIdentifierToken

	^type == #'identifier' and: [value = 'defined(' or: [value = 'defined']].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isDirectiveToken

	^type == #'directive'.
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isDotToken

	^type == #'punctuator' and: [value = '.'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isEmptyToken

	^type == #'empty'.
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isEndifDirective

	^type == #'directive' and: [value = 'endif'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isEqualsToken

	^type == #'punctuator' and: [value = '='].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isExtensionIdentifierToken

	^type == #'identifier' and: [value = '__extension__'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isHashToken

	^type == #'punctuator' and: [value = '#'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isIdentifierToken

	^type == #'identifier'.
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isIntegerToken

	^#(#'int16') includes: type.
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isOpenAngleBracketToken

	^type == #'punctuator' and: [value = '<'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isOpenCurlyBracketToken

	^type == #'punctuator' and: [value = '{'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isOpenParenthesisToken

	^type == #'punctuator' and: [value = '('].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isOpenSquareBracketToken

	^type == #'punctuator' and: [value = '['].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isSemicolonToken

	^type == #'punctuator' and: [value = ';'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isStarToken

	^type == #'punctuator' and: [value = '*'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isStringToken

	^type == #'string'.
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isStructIdentifierToken

	^type == #'identifier' and: [value = 'struct'].
%
category: 'Testing'
set compile_env: 0
method: CPreprocessorToken
isTildeToken

	^type == #'punctuator' and: [value = '~'].
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
key

	^type.
%
category: 'Accessing'
set compile_env: 0
method: CPreprocessorToken
line

	^line.
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
octalCharacterFrom: aStream
	"Enter with pointing to '0' of '\0ooo'"

	| string |
	string := String withAll: '8r0'.
	[true] whileTrue: [
		| char |
		char := self peekFrom: aStream.
		(string size < 6 and: [char isDigit]) ifFalse: [
			value add: (Character codePoint: (Integer fromString: string)).
			^self.
		].
		string add: char.
		source add: aStream next.
	].
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
peekFrom: aStream
	"Give back the next non-mergeLine token; does not skip whitespace, does not skip comments"

	| char1 char2 char3 |
	(char1 := aStream peek) isNil ifTrue: [^nil].
	char1 == $\ ifFalse: [^char1].
	(char2 := aStream peek2) isNil ifTrue: [^$\].
	(char2 == Character cr or: [char2 == Character lf]) ifFalse: [^$\].
	"$\ was followed by CR or LF"
	line := line + 1.
	source 
		add: aStream next;
		add: aStream next.
	(char3 := aStream peek) isNil ifTrue: [^nil].
	(char2 ~~ char3 and: [char3 == Character cr or: [char3 == Character lf]]) ifFalse: [^char3].
	^self peekFrom: aStream.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
printOn: aStream

	(source isNil or: [source isEmpty]) ifTrue: [
		type printOn: aStream.
		aStream nextPutAll: ' -> '.
		value printOn: aStream.
	] ifFalse: [
		aStream nextPutAll: source.
	].
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readCharacterFrom: aStream

	type := #'uint8'. 
	self 
		readLiteralUpTo: $'
		from: aStream.
	255 < (value := value first codePoint) ifTrue: [self error: 'Should be wide character!'].
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readHashFrom: aStream isBeginningOfLine: aBoolean

	| char |
	source add: aStream next.	"$#"
	(char := self peekFrom: aStream) = $# ifTrue: [	"## means to merge adjacent tokens"
		source add: aStream next.
		type := #'concatenation'.
		^self.
	].
	aBoolean ifFalse: [
		type := #'punctuator'.
		value := '#'.
		^self.
	].
	self 
		skipCommentsIn: aStream;
		readIdentifierFrom: aStream .
	type := #'directive'.
	value isEmpty ifTrue: [self error: 'Missing preprocessor directive'].
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readIdentifierFrom: aStream

	|  char |
	type := #'identifier'.
	value := String new.
	[
		char := self peekFrom: aStream.
		char notNil and: [char isAlphaNumeric or: [char = $_]].
	] whileTrue: [
		value add: aStream next.
	].
	"While not really part of an identifier, the open parenthesis helps us handle macro functions. See CPreprocessor>>#'macroExpansionOf:'"
	aStream peek == $( ifTrue: [value add: aStream next].
	source add: value.
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readLiteralUpTo: aCharacter from: aStream

	| char |
	source add: aStream next.	"The open quote character"
	value := String new.
	[
		(char := aStream next) == Character lf ifTrue: [self error: 'End of literal not found on current line!'].
		source add: char.
		char = aCharacter.
	] whileFalse: [
		char = $\ ifTrue: [
			source add: (char := aStream next).
			char = $a 	ifTrue: [value add: (Character codePoint: 7	)	] ifFalse: [		"\a 	= alert, <Ctrl>+<G>"
			char = $b 	ifTrue: [value add: (Character codePoint: 8	)	] ifFalse: [		"\b 	= backspace, <Ctrl>+<H>"
			char = $f	ifTrue: [value add: (Character codePoint: 12	)	] ifFalse: [		"\f 	= form feed, <Ctrl>+<L>"
			char = $n	ifTrue: [value add: (Character codePoint: 13	)	] ifFalse: [		"\n 	= new line, <Ctrl>+<M>"
			char = $r	ifTrue: [value add: (Character codePoint: 10	)	] ifFalse: [		"\r 	= carriage return, <Ctrl>+<J>"
			char = $t	ifTrue: [value add: (Character codePoint: 9	)	] ifFalse: [		"\t 	= horizontal tab, <Ctrl>+<I>"
			char = $v	ifTrue: [value add: (Character codePoint: 11	)	] ifFalse: [		"\v 	= vertical tab, <Ctrl>+<K>"
			char = $x	ifTrue: [self hexCharacterFrom: aStream			] ifFalse: [		"\xdd = hex number"
			char = $0	ifTrue: [self octalCharacterFrom: aStream		] ifFalse: [		"\0ddd	= octal number"
			char = $'	ifTrue: [value add: char									] ifFalse: [		"\'		= single quote"
			char = $"	ifTrue: [value add: char									] ifFalse: [		"""		= double quote"
			char = $?	ifTrue: [value add: char									] ifFalse: [		"?		= question mark"
			char = $\	ifTrue: [value add: char									] ifFalse: [		"\\		= single backslash"
			]]]]]]]]]]]]].
		] ifFalse: [
			value add: char.
		].
	].
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readNumberFrom: aStream
"A preprocessing number has a rather bizarre definition. The category includes all the normal integer and floating point constants one expects of C, but also a number of other things one might not initially recognize as a number. Formally, preprocessing numbers begin with an optional period, a required decimal digit, and then continue with any sequence of letters, digits, underscores, periods, and exponents. Exponents are the two-character sequences `e+', `e-', `E+', `E-', `p+', `p-', `P+', and `P-'. (The exponents that begin with `p' or `P' are new to C99. They are used for hexadecimal floating-point constants.)"

	| char string isLong isUnsigned |
	type := #'number'.
	value := String with: (source add: aStream next).
	[
		aStream atEnd not and: [
			char := self peekFrom: aStream.
			char isAlphaNumeric or: [char == $. or: [char == $_ or: [(char == $+ or: [char == $-]) and: ['eEpP' includes: value last]]]].
		].
	] whileTrue: [
		value add: (source add: aStream next).
	].
	isLong := false.
	isUnsigned := false.
	(2 < value size and: [(value at: 1) = $0 and: [(value at: 2) = $x]]) ifTrue: [
		string := '16r' , (value copyFrom: 3 to: value size).
	] ifFalse: [
		value first == $0 ifTrue: [
			string := '8r' , value.
		] ifFalse: [
			string := value.
		].
	].
	('lL' includes: string last) ifTrue: [
		isLong := true.
		string := string copyFrom: 1 to: string size - 1.
	].
	('uU' includes: string last) ifTrue: [
		isUnsigned := true.
		string := string copyFrom: 1 to: string size - 1.
		('lL' includes: string last) ifTrue: [
			isLong := true.
			string := string copyFrom: 1 to: string size - 1.
		].
	].
	value := Integer fromString: string.
	(value class ~~ SmallInteger and: [16rFFFFFFFFFFFFFFFF < value]) ifTrue: [self halt].
	type := (isLong or: [16rFFFFFFFF < value]) ifTrue: [
		(isUnsigned or: [16r7FFFFFFFFFFFFFFF < value]) 
			ifTrue: [#'uint64']
			ifFalse: [#'int64'].
	] ifFalse: [
		16rFFFF < value ifTrue: [
			(isUnsigned or: [16r7FFFFFFF < value]) 
				ifTrue: [#'uint32']
				ifFalse: [#'int32'].
		] ifFalse: [
			(isUnsigned or: [16r7FFF < value]) 
				ifTrue: [#'uint16']
				ifFalse: [#'int16'].
		].
	].
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readPunctuatorFrom: aStream

	| char1 char2 char3 twoChars threeChars |
	type := #'punctuator'.
	source add: (char1 := aStream next).
	value := String with: char1.
	(char2 := aStream peek) isNil ifTrue: [^self].
	(char3 := aStream peek2) isNil ifTrue: [
		(twoChars := value copy) add: char2 .
	] ifFalse: [
		(twoChars := value copy) add: char2 .
		(threeChars := twoChars copy) add: char3 .
	].
	(#('<<=' '>>=' '->*') includes: threeChars) ifTrue: [
		value := threeChars.
		source add: char2; add: char3.
		aStream next; next.
		^self.
	].
	(#('++' '--' '==' '!=' '<=' '>=' '&&' '||' '<<' '>>' '+=' '-=' '*=' '/=' '%=' '&=' '|=' '^=' '->' '.*' '::') includes: twoChars) ifTrue: [
		value := twoChars.
		source add: char2.
		aStream next.
	].
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readStringFrom: aStream

	type := #'string'. 
	self 
		readLiteralUpTo: $" 
		from: aStream.
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readWideCharacterFrom: aStream
		"#define __WCHAR_MAX__ 2147483647"

	(source add: aStream next) == $L ifFalse: [self error: 'How did we get here!?'].
	type := #'int32'.
	self 
		readLiteralUpTo: $'
		from: aStream.
	16r7FFFFFFF < (value := value first codePoint) ifTrue: [self error: 'Should be wide character!'].
%
category: 'Reading'
set compile_env: 0
method: CPreprocessorToken
readWideStringFrom: aStream

	(source add: aStream next) == $L ifFalse: [self error: 'How did we get here!?'].
	type := #'string'.
	self 
		readLiteralUpTo: $' 
		from: aStream.
	self halt.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
removeOpenParenthesis

	value size: value size - 1.
	source size: source size - 1.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
replaceFirst: aToken

	file := aToken file.
	line := aToken line.
	source := aToken source.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
replaceOther: aToken

	file := aToken file.
	line := aToken line.
	source := ''.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
setValueToArrayWithEmptyToken

	value := Array with: CPreprocessorToken empty.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
skipBlockCommentsIn: aStream

	| char |
	source add: aStream next.	"This should be the $* following the $/"
	[
		source add: (char := aStream next).
		char == Character lf ifTrue: [line := line + 1].
		char = $* and: [(self peekFrom: aStream) = $/].
	] whileFalse: [].
	source add: aStream next.	"This should be the $/ following the $*"
	^self skipCommentsIn: aStream.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
skipCommentsIn: aStream
		"answer true if came to end-of-line, otherwise false"

	| char position |
	[
		(char := aStream peek) notNil and: [char isSeparator].
	] whileTrue: [
		char == Character lf ifTrue: [^true].
		source add: aStream next.
	].
	char == $/ ifFalse: [^false].
	position := aStream position.
	source add: aStream next.
	char := self peekFrom: aStream.
	char = $* ifTrue: [^self skipBlockCommentsIn: aStream].
	char = $/ ifTrue: [^self skipLineCommentsIn: aStream].
	source size: source size - 1.
	aStream position: position.
	^false.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
skipLineCommentsIn: aStream

	source add: aStream next.	"This should be the $/ following the $/"
	[
		aStream peek == Character lf.
	] whileFalse: [
		source add: aStream next.
	].
	^true.
%
category: 'Accessing'
set compile_env: 0
method: CPreprocessorToken
source

	source isNil ifTrue: [^''].
	^source.
%
category: 'Accessing'
set compile_env: 0
method: CPreprocessorToken
type

	^type.
%
category: 'Accessing'
set compile_env: 0
method: CPreprocessorToken
value

	type == #'define' ifTrue: [self error: 'Should call #defineParameters or #defineString'].
	^value.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
value: anObject

	value := anObject.
%
category: 'other'
set compile_env: 0
method: CPreprocessorToken
valueString

	type == #'empty' ifTrue: [^''].
	type == #'identifier' ifTrue: [^value].
	type == #'punctuator' ifTrue: [^value].
	type == #'int16' ifTrue: [^value printString].
	self halt.
%
