!=========================================================================
! Copyright (C) VMware, Inc. 2008.  All Rights Reserved.
!
! $Id$
!
! File: Module.gs
!   
!=========================================================================

! Module created by bom.c

! should not need SELF_CAN_BE_SPECIAL, self should always be non-nil
!  in instances of Module

set class Module
! removeallmethods 
! removeallclassmethods 

category: 'Documentation'
classmethod: 
comment
^ '
 Module represents the Ruby  class Module .
 In a Smalltalk repository, there are no instances of Module and
 creating instances of Module is not supported.
 
   InstVars
      name             a Symbol
      classHistory     nil for Ruby , in a Smalltalk Class nil or a ClassHistory 
      transientMethDicts   always nil on disk, in memory, nil or an Array .
        transient methodDict for envId is      transientMethDicts at:(envId*4 + 1)
        transient rubyNameSpace for envId is   transientMethDicts at:(envId*4 + 2)
        transient rubySuperclass for envId is  transientMethDicts at:(envId*4 + 3)
 '
%

! instance creation not supported in Smalltalk image , see newModule in .mcz

category: 'Smalltalk Printing'
classmethod:
sprintf: formatString with: argsArray

"Ruby  printf per Pickaxe book page 529..532  .
  this primitive is used from Ruby as instance method in Kernel,
  and is here for printing from Smalltalk. "
<primitive: 767>
formatString _validateClass: String .
argsArray _validateClass: Array  .
self _primitiveFailed: #sprintf:with:
%

! --------------------
category: 'Accessing'
method:
name
  ^ name
%

method:
_name
  ^ name
%

method:
name: aName
  self isInvariant ifTrue:[ 
     ArgumentTypeError signal:'illegal attempt to change name of a Module'.
  ].
  name := aName asSymbol
%

method:
classHistory
  "Modules have a classHistory instVar, but it is not used"
  ^ nil
%

method:
transientMethodDictForEnv: envId
"result will be nil if no methods exist for specified environmentId.
 Does not  include ruby auto-init from primaryCopy. "

^ transientMethDicts atOrNil: (envId*4 + 1) .
%

method:
transientMethodDictForEnv: envId put: aValue
  "aValue should be a GsMethodDictionary, or nil ,
   caller responsible for _refreshClassCache "

<protected>
| ofs |
ofs := envId*4 + 1 .
transientMethDicts size < ofs ifTrue:[ transientMethDicts size: ofs ].
transientMethDicts at: ofs put: aValue
%


method:
rubySuperclass: envId

"for receiver and specified environment, return first non-nil of:
    transient super class
    persistent super class
    env 0 super class (smalltalk superclass) "

<primitive: 676>
envId _validateClass: SmallInteger .
(envId < 0 or:[ envId > 255]) ifTrue:[ "COMPILE_ENV_MAX"
  self _errorIndexOutOfRange: envId 
].
self _primitiveFailed: #rubySuperclass:
%

! --------------------
category: 'Queries'
method: 
isMeta

"Returns whether the receiver is a kind of metaclass."

^ false
%

method: 
thisClass

"Returns the receiver.  
 For a Class the receiver's metaclass returns this Class as well.
 This method is useful to get the base version of a Class if one is
 holding either the Class or its metaclass."

^ self
%

! --------------------
category: 'Authorization'
method:
objectSecurityPolicy: anObjectSecurityPolicy

"Assigns the receiver and its non-shared components to the given security policy.
 The security policies of class variable values are not changed.  The current user 
 must have write access to both the old and new security policies for this method 
 to succeed."

  | anArray aCls categBlk |

  self lockableParts accompaniedBy: anObjectSecurityPolicy do: [:policy :each |
    "categorys, methDicts   instVars for  self  and  self class  
     handled here.  For Class, lockableParts includes extraDict. "
    each _objectSecurityPolicy: policy  
  ].
  anArray := { } .
  aCls := self .
  categBlk := [:arr | 
    arr ifNotNil:[ 
      anArray add: arr .
      arr _isArray ifTrue:[ 
        anArray addAll: arr . arr do:[:dict | anArray addAll: dict values].
      ] ifFalse:[
        anArray addAll: arr values
      ].
    ]
  ].
  2 timesRepeat:[ | arr |
    aCls persistentMethodDictsDo:[ :aDict | anArray add: aDict ; addAll: aDict values ].
    aCls persistentNameSpacesDo:[ :aRns | anArray add: aRns ].
    categBlk value: aCls __categorys .
    aCls _poolDictionaries do:[:d | anArray add: d ].
    aCls := self class .
  ].
  anArray add: classHistory.
  constraints _isArray ifTrue: [ anArray add: constraints ].
  (anArray asIdentitySet) accompaniedBy: anObjectSecurityPolicy do: [ :policy :anObj | 
    ( anObj isSpecial or: [ anObj _isSymbol or:[ anObj == #() ]] ) ifFalse:[ 
       anObj _objectSecurityPolicy: policy 
    ].
  ].
%

category: 'Locking'
method: 
lockableParts

"Returns an Array of the receiver's contents that are locked by browsers
 and folders."

| mcls parts |
mcls := self class .
parts := { self . 
  instVarNames . 
  classVars .  "if not nil will be shared with the metaClass"
  categorys . 
  methDicts . 

  mcls . 
  mcls.instVarNames . 
  mcls.categorys . 
  mcls.methDicts . 
  mcls.poolDictionaries }.
parts := parts reject:[:ea | ea == nil ].
^parts
%

category: 'Modifying Classes'
method:
_setClassVars: aDict
  classVars ifNil:[ 
    classVars := aDict .
  ] ifNotNil:[ aDict ~~ classVars 
		  ifTrue:[ self error:'invalid store to classVars']
  ].
%
!---------------------------------
category: 'Browser Methods'

method: 
changeNameTo: aString
  "Sets the receiver's 'name' instance variable."

  self _validatePrivilege ifTrue:[
    aString size == 0 ifTrue: [ ^ self error: 'class names must have size > 0'].
    name := Symbol withAll: aString.
  ]
%

method: 
hierarchy

"Returns a String that enumerates the receiver's superclasses (up to Object)
 and the instance variables defined by the receiver and each of its
 superclasses."

^ self hierarchy: 0
%

method:
hierarchy: envId

"Returns a String that enumerates the receiver's superclasses (up to Object)
 and the instance variables defined by the receiver and each of its
 superclasses."

^ self _superClassHierarchy: envId includeOops: false includeVirtual: false
%

method: 
rubyHierarchy: envId
"Returns a String that enumerates the receiver's superclasses (up to Object)
 and the instance variables defined by the receiver and each of its
 superclasses. Includes virtual classes"

^ self _superClassHierarchy: envId includeOops: false includeVirtual: true
%

method: 
_superClassHierarchy: envId includeOops: withOopsBool includeVirtual: vBool

"Returns a String showing the superClass hierarchy of the receiver.
 Used by implementation of topaz. "

| lf depth clsList rptList result |
lf := Character lf.
depth := 0.
(clsList := self _allSuperList: vBool env: envId ) addLast: self .
rptList := { }  .
depth := 0 .
clsList do:[ :aCls | | aLine ivNames |
  aLine := String new .
  1 to: depth do: [:i | aLine addAll: '  '].
  aLine add: (aCls name  ifNil:[ '(nil name)' ]) .
  ivNames := aCls _instVarNamesWithSeparator:  ''   .
  ivNames size ~~ 0 ifTrue: [
     aLine add: $(; add: ivNames ; add: $) .
  ].
  depth := depth + 1 .
  rptList add: aLine .
].
withOopsBool ifTrue:[ | width padStr |
  width := 0 .
  rptList do:[:aLine | | sz |
    ((sz := aLine size) > width and:[ sz < 40 ]) ifTrue:[ width := sz ]
  ].
  width := width + 4 .
  padStr := String new .
  width timesRepeat:[ padStr add: $  . ].
  1 to: rptList size do:[:idx | | aCls priCls aLine lsz |
    aLine := rptList at: idx .
    lsz := aLine size .
    lsz < width ifTrue:[
      aLine size: width .
      aLine replaceFrom: lsz + 1 to: width with: padStr startingAt: lsz + 1 .

    ] ifFalse:[  aLine add: $  ].
    aLine add: (aCls := clsList at: idx) asOop asString .
    vBool ifTrue:[ | ns |
      "this path only used for Ruby"
      aLine add: '  format=16r'; add: (aCls format asHexString) .
      priCls := aCls rubyPrimaryCopy .
      priCls ~~ aCls ifTrue:[
        aLine add: ' primaryCopy '; add: priCls asOop asString .
      ].
      "persistentNameSpace:, transientNameSpace: implemented in Maglev*.mcz"
      ns := aCls persistentNameSpace: envId .
      aLine add: ' pns=' ; add: ( ns ifNil:[ 'nil'] ifNotNil:[ ns asOop asString ]).
      ns := aCls transientNameSpace: envId .
      aLine add: ' tns=' ; add: ( ns ifNil:[ 'nil'] ifNotNil:[ ns asOop asString ]).
    ].
  ].
].
result := String new .
rptList do:[ :aLine | result add: aLine ; add: lf .].
^ result
%
method: 
_instVarNamesWithSeparator: sep

"Returns a string showing my instance variables, with the given
 separator string inserted after every three names."

| result i theIvs numIvs each |

result := String new: 0.
i := 0.
theIvs := self instVarNames.
numIvs := theIvs size.
1 to: numIvs do:[:j|
  each := theIvs at: j .
  (i := i + 1) > 3 ifTrue:
    [
    result addAll: sep.
    i := 0.
    ].
  result addLast: $  .
  (each includesValue: $') 
    ifTrue:[ result addAll: each _asSource ]
    ifFalse:[ result addAll: each ].
  ].
^result
%

!--------------
category: 'Browsing'
method:
definition

  InternalError new ; 
    details:'Unexpected instance of Module in a Smalltalk repository' ; 
    signalNotTrappable  .

  ^ 'Module newModule name: ', name printString , '
       "does not include code to install in a dictionary" '
%
method:
ogDefinition
  "squeak browser support"
  ^ self definition
%
method:
allSubclassesDo: aBlock
 ^ self
%
method:
allSubclasses
  ^ { }
%
method:
subclasses
  ^ { }
%

!-----------------
category: 'Clustering'
method: 
clusterBehavior

 super clusterBehavior ifTrue:[ ^ true ].
 classHistory ifNotNil:[ classHistory cluster ].
 ^ false
%
