!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: ProfMonitorEntry.gs 25988 2011-06-14 17:55:54Z stever $
!
! Superclass Hierarchy:
!   ProfMonitorEntry, Object.
!
!=========================================================================

expectvalue %String
run
| oldClass expectedInstVars newClass |
expectedInstVars := #(cmethod tally parents
                    children total rcvrClass recursed
                    childTallies parentTallies totalSamples elapsedTime).
oldClass := Globals at: #ProfMonitorEntry otherwise: nil.
(oldClass ~~ nil and: [oldClass instVarNames = expectedInstVars]) ifTrue: [
	^'Found expected class'.
].
newClass := Object 
  subclass: 'ProfMonitorEntry'
  instVarNames: expectedInstVars
  classVars: #()
  classInstVars: #()
  poolDictionaries: { }
  inDictionary: Globals .
^(oldClass ifNil: ['Created new class:'] 
	ifNotNil: ['Replaced old class with:']) , '
' , newClass definition.

%

! remove existing behavior from ProfMonitorEntry
removeallmethods ProfMonitorEntry
removeallclassmethods ProfMonitorEntry
set class ProfMonitorEntry

! ------------------- Class methods for ProfMonitorEntry
category: 'Documentation'
classmethod:
comment
^'
 A ProfMonitorEntry represents one stack frame of a sample
 taken by a ProfMonitor instance.'
%

category: 'Instance Creation'
classmethod: ProfMonitorEntry
newForMethod: aMethod receiverClass: rcvrClass

  | pme |
  aMethod ifNil:[ ^ self error:'method must be non-nil'].
  rcvrClass ifNil:[ ^ self error:'rcvrClass must be non-nil'].
  aMethod class == GsNMethod ifFalse:[ ^ self error:'not a method'].
  pme := self _basicNew.
  pme cmethod: aMethod.
  pme tally: 0 .
  pme total: 0 .
  pme children: { } .
  pme rcvrClass: rcvrClass.
  ^pme.
%

! ------------------- Instance methods for ProfMonitorEntry
category: 'Accessing'
method: 
children

   "Return the value of the instance variable 'children'."
   ^children
%
category: 'Updating'
method: 
children: newValue

   "Modify the value of the instance variable 'children'."
   children := newValue
%
category: 'Accessing'
method: 
childTallies

   "Return the value of the instance variable 'childTallies'."
   ^childTallies
%
category: 'Updating'
method: 
childTallies: newValue

   "Modify the value of the instance variable 'childTallies'."
   childTallies := newValue
%
category: 'Accessing'
method: 
cmethod

   "Return the value of the instance variable 'cmethod'."
   ^cmethod
%
category: 'Updating'
method: 
cmethod: newValue

   "Modify the value of the instance variable 'cmethod'."
   cmethod := newValue
%
category: 'Accessing'
method: 
elapsedTime

   "Return the value of the instance variable 'elapsedTime'."
   ^elapsedTime
%
category: 'Updating'
method: 
elapsedTime: newValue

   "Modify the value of the instance variable 'elapsedTime'."
   elapsedTime := newValue
%
category: 'Accessing'
method: 
parents

   "Return the value of the instance variable 'parents'."
   ^parents
%
category: 'Updating'
method: 
parents: newValue

   "Modify the value of the instance variable 'parents'."
   parents := newValue
%
category: 'Accessing'
method: 
parentTallies

   "Return the value of the instance variable 'parentTallies'."
   ^parentTallies
%
category: 'Updating'
method: 
parentTallies: newValue

   "Modify the value of the instance variable 'parentTallies'."
   parentTallies := newValue
%
category: 'Accessing'
method: 
rcvrClass

   "Return the value of the instance variable 'rcvrClass'."
   ^rcvrClass
%
category: 'Updating'
method: 
rcvrClass: newValue

   "Modify the value of the instance variable 'rcvrClass'."
   rcvrClass := newValue
%
category: 'Accessing'
method: 
recursed

   "Return the value of the instance variable 'recursed'."
   ^recursed
%
category: 'Updating'
method: 
recursed: newValue

   "Modify the value of the instance variable 'recursed'."
   recursed := newValue
%
category: 'Accessing'
method: 
tally

   "Return the value of the instance variable 'tally'."
   ^tally
%
category: 'Updating'
method: 
tally: newValue

   "Modify the value of the instance variable 'tally'."
   tally := newValue
%
category: 'Accessing'
method: 
total

   "Return the value of the instance variable 'total'."
   ^total
%
category: 'Updating'
method: 
total: newValue

   "Modify the value of the instance variable 'total'."
   total := newValue
%
category: 'Accessing'
method: 
totalSamples

   "Return the value of the instance variable 'totalSamples'."
   ^totalSamples
%
category: 'Updating'
method: 
totalSamples: newValue

   "Modify the value of the instance variable 'totalSamples'."
   totalSamples := newValue
%

category: 'Reporting'

! reimplemented in Maglev.mcz 
method:
asStringWidth: nameWidth
  | str mth |
  mth := cmethod .
  mth == #GCI ifTrue:[
    str := #GCI asString width: nameWidth .
  ] ifFalse:[ | mthCls rcvCls |
    str := mth _classAndSelectorNameWidth: nameWidth .  
    (mth inClass ~~ (rcvCls := rcvrClass) and:[ rcvCls ~~ Object]) ifTrue:[
      str add: ' [' ; add: rcvCls name ; add: ']' .
    ].
  ].
  ^ str
%
