!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: ZeroDivide.gs 25988 2011-06-14 17:55:54Z stever $
!
!=========================================================================

set class ZeroDivide

! no removeall 
! ZeroDivide>>initialize is in Exception.gs

category: 'For Documentation Installation only'
classmethod: 
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'The class protocol describe the behavior of the global ZeroDivide. 
It is used to as an exception selector to catch zero divide exceptions 
and can also be used to signal that a division by zero error has 
occured. Zero divide exceptions are resumable so any message in this 
protocol that signal such an exception may ultimately return to their 
sender. The signaled exceptions generated by this type of object 
conform to the protocol <ZeroDivide>.

The instance protocol describes the behavior of exceptions that are 
signalled when an attempt is made to divide some number (the dividend) 
by zero.'.
doc documentClassWith: txt.

self description: doc.
%

! -----------------------------------------
category: 'Instance creation'
classmethod: 
dividend: argument
	"Signal the occurrence of a division by zero exception. Capture the number
	that was being divided such that it is available from the 'signaled exception.'

	If the message #dividend is subsequently sent to the <ZeroDivide> object
	that is the 'signaled exception' the value of 'argument' is returned."

	"It seems rare that this Exception would be signaled by sending this message.
	Typically, this Exception is signaled indirectly by the VM when a 
	division by zero occurs. So why not just take the usual path?!"

	argument / 0.
%

classmethod: 
signal

	0 / 0.
%

! -----------------------------------------
category: 'Accessing'
method: 
dividend

	^ dividend
%

method: 
dividend: aNumber

  dividend := aNumber
%

! fixed 41582
method: 
asString
^ self _description: 
   'attempt to divide ' , dividend asString , ' by zero' 
%
