!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: rcrangeindex.gs 25988 2011-06-14 17:55:54Z stever $
!
! Superclass Hierarchy:
!   RcRangeEqualityIndex, RangeEqualityIndex, IdentityIndex, Array, SequenceableCollection,
!   Collection, Object.
!
! class created in idxclasses.topaz
!=========================================================================

removeallmethods RcRangeEqualityIndex
removeallclassmethods RcRangeEqualityIndex

category: 'For Documentation Installation only'
classmethod: RcRangeEqualityIndex
installDocumentation

| doc txt |
doc := GsClassDocumentation _newForPrivateGsClass: self.

txt := (GsDocText new) details:
''.

doc documentClassWith: txt.

self description: doc.
%

! ------------------- Class methods for RcRangeEqualityIndex
category: 'Instance Creation'
classmethod: RcRangeEqualityIndex
newWithLastElementClass: aClass

"Create a new instance and initialize its B-tree root."

| newOne |
newOne := super new.
newOne lastElementClass: aClass.
newOne btreeRoot: aClass rcBtreeLeafNodeClass new.
^ newOne
%
! ------------------- Instance methods for RcRangeEqualityIndex
category: 'Updating'
method: RcRangeEqualityIndex
btreeAt: aKey put: aValue

"Insert the key/value pair into the root B-tree node.  Must check to see if a
 split occurred.  If so, create a new root node and put the old root and the
 new split node into the new root."

(self _canCompareWith: aKey)
    ifFalse: [
      ^ self _error: #rtErrRangeEqualityIndexInvalidClassKindForBtree
        args: { aKey . lastElementClass }
    ].

self _btreeAt: aKey put: aValue logging: true
%
category: 'Updating'
method: RcRangeEqualityIndex
btreeRemoveKey: aKey value: aValue

"Removes the key and value from the B-tree root node.  Must check to see if
 a merge occurred in the root such that the root only contains a single
 entry.  If so, make the single entry the new root.  Returns whether the
 removal occurred."

^ self _btreeRemoveKey: aKey value: aValue logging: true
%
category: 'Updating'
method: RcRangeEqualityIndex
_btreeAt: aKey put: aValue logging: aBoolean


        ^self _btreeAt: aKey put: aValue logging: aBoolean selectiveAbortSet: nil
%
category: 'Updating'
method: RcRangeEqualityIndex
_btreeAt: aKey put: aValue logging: aBoolean selectiveAbortSet: selectiveAbortSetOrNil

"Insert the key/value pair into the root B-tree node.  Must check to see if a
 split occurred.  If so, create a new root node and put the old root and the
 new split node into the new root."

| returnNode node |
returnNode := btreeRoot btreeAt: aKey put: aValue for: self selectiveAbortSet: selectiveAbortSetOrNil.

" see if a split occurred "
returnNode == btreeRoot
    ifFalse: [ " returnNode is the second half of the split "

          selectiveAbortSetOrNil ~~ nil ifTrue: [ selectiveAbortSetOrNil add: returnNode ].
        System _addRootObjectToRcReadSet: returnNode.
        System redoLog addConflictObject: returnNode for: self.

        " create the new parent node "
        node := btreeRoot parentNodeClass new.
        node lastValue: returnNode lastValue.

        " insert the first half (the original root) "
        node
            _insertKey: (btreeRoot _at: btreeRoot _lastKeyIndex)
            value: btreeRoot
            atIndex: 1.
        " insert the second half "
        node
            _insertKey: (returnNode _at: returnNode _lastKeyIndex)
            value: returnNode
            atIndex: (node entrySize + 1).

          selectiveAbortSetOrNil ~~ nil ifTrue: [ selectiveAbortSetOrNil add: node ].
        System _addRootObjectToRcReadSet: node.
        System redoLog addConflictObject: node for: self.

        btreeRoot := node.
    ].
System _addRootObjectToRcReadSet: self.
aBoolean 
  ifTrue: [ 
    self _logBtreeAt: aKey put: aValue. 
  ].
^true
%
category: 'Updating'
method: RcRangeEqualityIndex
_btreeRemoveKey: aKey value: aValue logging: aBoolean

        ^self  _btreeRemoveKey: aKey value: aValue logging: aBoolean selectiveAbortSet: nil
%
category: 'Updating'
method: RcRangeEqualityIndex
_btreeRemoveKey: aKey value: aValue logging: aBoolean selectiveAbortSet: selectiveAbortSetOrNil

"Removes the key and value from the B-tree root node.  Must check to see if
 a merge occurred in the root such that the root only contains a single
 entry.  If so, make the single entry the new root.  Returns whether the
 removal occurred."

(btreeRoot removeKey: aKey value: aValue for: self selectiveAbortSet: selectiveAbortSetOrNil)
    ifTrue: [
        ((btreeRoot numElements == 1) and: [ btreeRoot isLeaf not ] )
            ifTrue: [ btreeRoot := btreeRoot at: (btreeRoot _lastKeyIndex - 1) ].
          System _addRootObjectToRcReadSet: self.
        aBoolean 
          ifTrue: [ 
            self _logBtreeRemoveKey: aKey value: aValue.
          ].
        ^ true
    ]
    ifFalse: [ ^ false ].
%
category: 'Private'
method: RcRangeEqualityIndex
_logBtreeAt: aKey put: aValue

"Create a log entry for adding the given key/value."

| logEntry |
logEntry := LogEntry new.
logEntry receiver: self;
    selector: #_btreeAt:put:logging:selectiveAbortSet:;
    argArray: { aKey . aValue . false . nil }.
System redoLog addLogEntry: logEntry
%
category: 'Private'
method: RcRangeEqualityIndex
_logBtreeRemoveKey: aKey value: aValue

"Create a log entry for removing the given key/value."

| logEntry |
logEntry := LogEntry new.
logEntry receiver: self;
    selector: #_btreeRemoveKey:value:logging:selectiveAbortSet:;
    argArray: { aKey . aValue . false . nil }.
System redoLog addLogEntry: logEntry
%
category: 'Private'
method: RcRangeEqualityIndex
_resolveRcConflictsWith: conflictObjects

"A logical write-write conflict has occurred on the receiver.  The objects that
 had the actual physical write-write conflicts are in the conflictObjects
 Array.  Selectively abort the receiver and then attempt to replay the
 operations maintained in the System redo log.  Returns true if all the
 operations could be replayed; otherwise returns false."

"inline implementation of _abortAndReplay: to abort old btrees before
 replaying log entries."

| redoLog logEntries selectiveAbortSet |
redoLog := System _redoLog.
    
" if no log entries to replay, then we're done "
redoLog == nil ifTrue: [ ^ false ].
logEntries := redoLog getLogEntriesFor: self .
logEntries == nil ifTrue:[ ^ false ].
   
" cannot perform selective abort if receiver has a dependency tag "
self _hasDependencyList ifTrue: [ ^ false ].

" explicitly abort the btrees that made up original changes, since we need to abort entire tree, before replaying, since
  nodes may split differently."
selectiveAbortSet := IdentitySet new.
1 to: conflictObjects size do:[ :j | | co |
  co := conflictObjects at: j.
  co _selectiveAbort.
  selectiveAbortSet add: co.
].
    
"abort all of the btree nodes involved in the transaction"
redoLog conflictObjects keysAndValuesDo:[ :conflictObject :redoObject |
  redoObject == self 
        ifTrue: [ 
                conflictObject _selectiveAbort.
                selectiveAbortSet add: conflictObject ].
].

" refresh the state of the receiver "
self _selectiveAbort.
    
" replay all operations on the receiver"
1 to: logEntries size do:[:j | | logEnt args |
  logEnt := logEntries at: j .
  logEnt argArray at: 4 put: selectiveAbortSet. "install the selectiveAbortSet into the #selectiveAbortSet arg slot"
  logEnt redo ifFalse:[
    ^ false  "redo failed, cannot commit"
  ].
].
^ true
%
