!=========================================================================
! Copyright (C) VMware, Inc. 2008-2011.  All Rights Reserved.
!
! $Id: GsProcess_ruby.gs 25988 2011-06-14 17:55:54Z stever $
!
!  additional methods  for GsProcess to support Ruby 
!=========================================================================

set class GsProcess
! adding additional methods , don't 'removeallmethods'

category: 'Ruby support'

!Allow subclasses of GsProcess for Ruby
expectvalue true
run
GsProcess allowSubclasses .
^ GsProcess subclassesDisallowed == false
%


method:
rubyPrivateSize
^ 20 "inline  GsProcess instSize"
%

classmethod: 
rubyPrivateInstSize
^ GsProcess instSize 
%

classmethod:
_rubyThreadDataAt: anOffset
 " Returns the Ruby thread-local data specified by anOffset, or nil.

  anOffset 	data
    1           result of RubyCompilerState current
    2           Ruby global $? , last child process status (a SmallInteger)
    3           Ruby instance_eval self's stack
    4           IdentitySet for ruby RecursionGuard
   Offsets symbolically named see *RubyGsProcessClientData* in opalcls.ht, bom.c
 "

 ^ self _current _rubyThreadDataAt: anOffset
%

method:
_rubyThreadDataAt: anOffset
 " Returns the Ruby thread-local data specified by anOffset, or nil.

  anOffset 	data
    1           result of RubyCompilerState current
    2           Ruby global $? , last child process status (a SmallInteger)
    3           Ruby instance_eval self's stack
    4           IdentitySet for ruby RecursionGuard
   Offsets symbolically named see *RubyGsProcessClientData* in opalcls.ht, bom.c
 "

 | thrData |
 thrData := clientData .
 thrData ifNil:[
   ^ nil .
 ].
 ^ thrData at: anOffset
%

classmethod:
_recursionGuardSet
 | s curr |
 s := (curr := self _current) _rubyThreadDataAt: GC_RubyGsProcessClientData_recursionGuardSet .
 s ifNil:[
   curr _rubyThreadDataAt: GC_RubyGsProcessClientData_recursionGuardSet put: ( s := IdentitySet new) .
 ].
 ^ s 
%

classmethod:
_rubyThreadDataAt: anOffset put: aValue

 "Returns receiver.  Store aValue into Ruby thread-local data.
  anOffset defined by  comments in _rubyThreadDataAt: "

 ^ self _current _rubyThreadDataAt: anOffset put: aValue
%

method:
_rubyThreadDataAt: anOffset put: aValue
 | thrData |
 (thrData := clientData ) ifNil:[
   thrData := Array new: OC_RubyGsProcessClientDataSize .
   clientData := thrData .
 ] .
 thrData at: anOffset put: aValue
%

method:
rubyPrivateSize
  ^ GsProcess instSize
%
classmethod:
rubyPrivateInstSize
  ^ GsProcess instSize
%

! initRubyMainThread moved to .mcz

classmethod:
allProcesses

  ^ self _scheduler allProcesses asArray
%

classmethod:
rubyMain
  "The RubyMainThread is initialized in class RubyCompiler.  "

  ^ SessionTemps current at: #RubyMainThread otherwise: nil
%  

classmethod:
pass

  "Returns receiver."
  self _scheduler yield
%

classmethod:
stop
  		"TODO set Thread.critical to false"
  self _current suspend .
%  

method:
threadDataAt: aName
  | dict |
  aName _isSymbol ifTrue:[
    (dict := environment) ifNil:[ ^ nil ].
    ^ dict at: aName otherwise: nil .
  ] ifFalse:[
    aName _isOneByteString ifFalse:[
      ArgumentError signal:'argument is not a String or Symbol'.
    ].
    (dict := environment) ifNil:[ ^ nil ].
    ^ dict at: aName asSymbol otherwise: nil .
  ]
%

method:
threadDataAt: aName put: aValue
  "Returns aValue"
  | dict |
  aName _isOneByteString ifFalse:[
    ArgumentError signal:'argument is not a String or Symbol'.
  ].
  (dict := environment) ifNil:[ 
     dict := IdentityKeyValueDictionary new .
     environment := dict 
  ].
  dict at: aName asSymbol put: aValue .
  ^ aValue
%

method:
keys
 | dict |
 (dict := environment) ifNil:[ ^ { } ].
 ^ dict keysAsArray 
%

method:
includesKey: aName
  | dict |
  aName _isOneByteString ifFalse:[
    ArgumentError signal:'argument is not a String or Symbol'.
  ].
  (dict := environment) ifNil:[ ^ false ].
  ^ (dict at: aName asSymbol otherwise: nil) ~~ nil 
%


classmethod:
exit
  | sched nextProc curr |
  sched := self _scheduler .
  curr := self _current .
  [
    curr _terminationStarted ifFalse:[ 
      curr _setModeinfoBit: ModeInfo_terminationStarted .
      curr _executeEnsureBlocks
    ]. 
  ] onException: AbstractException do:[ :ex |
    "ignore exception"
  ].
  curr _finishTermination . 
  nextProc := sched _findReadyProcessForExit: curr .
  nextProc ifNil:[  "no other processes to run"
    RubySystemExit signal:'Thread.exit called' 
  ] ifNotNil:[
    sched _switchFrom: nil to: nextProc .
  ].
  self _uncontinuableError . "should not reach here"
%  

method:
exit
 | curr |
 curr := self _current .
 (self == curr or:[ self == GsProcess rubyMain]) ifTrue:[ ^ GsProcess exit ].

 self _terminationStarted ifTrue:[ ^ self ].
 
 self terminate .
%

method:
alive

^ self _isTerminated not
%
 
method:
rubyGroup
  "Return the process group(a RubyThreadGroup ) the receiver is in.
   Assign receiver to the default group if it has no group."

  | g |
  (g := group) ifNil:[ 
    g := RubyThreadGroup default .
    group := g
   ].
  ^ g
%

method:
_start
  "Called from C to start a new process. 
   NOTE, exception handling differs in extent0.dbf and extent0.ruby.dbf .
   This version of _start from extent0.ruby.dbf .

   This method will never return. Once the receiver completes it should
   find another process to run. 
   This method is preloaded during VM startup.  You must logout/login
   to have any changes to this method take effect."
  | res envId curr |
  (modeInfo bitAnd: ModeInfo_continuationMask) ~~ 0 ifTrue:[
    self error:'cannot start an instance that is a continuation'
  ].
  block ifNotNil:[
    res :=
    (envId := msgEnvironmentId) == 0 ifTrue:[ 
      [ 
        block valueWithArguments: args 
      ] onException: { RubyThrowException . RubyBreakException . RubySystemExit }
        do: [ :ex | 
               "unexpected Ruby exception in a Smalltalk GsProcess."
               ex copy signalNotTrappable  .  " ... to GCI"
               ex resume .  "for GciContinue"
       ].
    ] ifFalse:[ 
      self _setModeinfoBit: ModeInfo_isRubyThread .
      self _startRuby: envId
    ].
  ].
  "if this process resumed a continuation, then current process may
   no longer be self at this point."
  curr := self _terminateCurrentWithResult: res . "terminate currently running process"
  joiners ifNotNil:[ 
    curr == self ifTrue:[ self _signalJoiners ]
               ifFalse:[ self error:'cannot _signalJoiners after resuming continuation'].
  ].
  self _scheduler _runNextProcess .
%


method:
joinValue
  self _join: 1000000000 .
  ^ blockResult .
%

method:
_join: limitSeconds
  | cls res |
  res := self join: limitSeconds .
  res == nil ifTrue:[ ^ nil "timed out" ].
  res := blockResult .
  (cls := res class) _isExceptionClass ifTrue:[ 
    cls == CannotReturn ifTrue:[ 
      ThreadError signal:'return cannot be used to exit a Thread'
    ].
    res copy signal 
  ].
  ^ self
%

method:
rubyPriority: anInt

anInt _isSmallInteger ifFalse:[
    ArgumentError signal:'expected a Fixnum' .
].
self priority: anInt
%

method:
rubyStatus

  self == self _current ifTrue:[ ^ 'run' copy ].
  self _isTerminated ifTrue:[
    blockResult == _remoteNil ifTrue:[ ^ nil ].
    ^ false
  ].
  ^ 'sleep' copy
%

method:
rubyStopped
  ^ (self _current) ~~ self
%

classmethod
rubyBasicNew

^ self rubyBasicNew_stBaseClass: GsProcess
%

classmethod:
rubyStart: argsArray block: aBlock
  "a ruby primitive"
  | thr envId sched |
  envId := self __callerEnvId .
  thr := (self rubyBasicNew_stBaseClass: GsProcess) 
           _init: aBlock args: argsArray env: envId 
              stackSerial: self _current _stackSerialNum 
	      forked:  ModeInfo_forked .
  (sched := self _scheduler) _scheduleNewProcess: thr .
  sched yield .
  ^ thr
%

method:
rubyStart: argsArray block: aBlock
  "a ruby primitive"
  | envId sched currProc |
  envId := self __callerEnvId .
  (currProc := self _current) == self ifTrue:[
    InternalError new 
      details:'GsProcess>>rubyStart, inconsistent current process' ;
      signalNotTrappable
  ].
  self _init: aBlock args: argsArray env: envId 
              stackSerial: currProc  _stackSerialNum 
              forked:  ModeInfo_forked .
  (sched := self _scheduler) _scheduleNewProcess: self .
  sched yield .
  ^ self
%

method:
rubyRun
  self rubyResume .
  self _scheduler yield
%

method:
rubyResume
  (modeInfo bitAnd: ModeInfo_anyTermination) ~~ 0 ifTrue:[
    ThreadError signal:'cannot resume a terminated thread'
  ].
  ^ self resume
%

method:
rubyPriority
  ^ self priority - 15 .
%

method:
rubyPriority: anInt
  anInt _isSmallInteger ifFalse:[
    ArgumentError signal:'expected a Fixnum'
  ].
  self priority: anInt + 15
%
method:
clearIoSelectResult
  | res |
  res := ioSelectResult .
  ioSelectResult := nil .
  ^ res
%

category: 'Private Ruby Scheduling'
method:
_reapEvents
  waitingOn := nil . 
  self _scheduler _scheduleProcess: self
% 

method:
_setTerminated
  "for use in RubySystemExit processing only"

  self _setModeinfoBit: ModeInfo_terminated
%

